function statistics = computeStatistics(dens,grid,quantiles,z)

% Computes certain summary statistics of a density which is available on a
% grid.
%
% See: copulaPredictDensity, copulaCVDensity
%
% Inputs:
%
% dens.........(predictive) density computed on a grid e.g. by
%              copulaPredictDensity or copulaCVDensity
%
% grid.........the grid on which the density is available
%
% quantiles....the vector of quantiles to compute
%
% z............observations (z-values, data.z) which are used to compute PIT
%              and LS values
%
% Outputs:
%
% statistics...a structure array of summary statistics
%              statistics.mean...the mean of the distribution
%              statistics.std...the standard deviation of the distribution
%              statistics.modal...the mode of the distribution
%              statistics.median...the median of the distribution 
%              statistics.quantiles...the quantiles of the distribution if
%                                     the input vector "quantiles" is nonempty
%              statistics.PIT...Probability Integral Transform values if
%                               input "z" is nonempty
%              statistics.LS...Logarithmic Score if input "z" is nonempty.
%  
% Hannes Kazianka, August 2011
    
    if nargin<4
        z=[];
    end
    if nargin<3
        quantiles=[];
    end
    
    n=size(dens,1);
    quant=NaN*ones(n,length(quantiles));
    median=NaN*ones(n,1);
    modal=NaN*ones(n,1);
    mean1=NaN*ones(n,1);
    std=NaN*ones(n,1);
    PIT=NaN*ones(n,1);
    LS=NaN*ones(n,1);

    dens=diag(1./sum(dens,2))*dens;
    for i=1:n
        cumdistribfunction=cumsum(dens(i,:));
        indmedian=find(cumdistribfunction<=0.5,1,'last');
        if isempty(indmedian)
            median(i)=grid(1);
        else
            
        median(i)=grid(indmedian);
        end
        modal(i)=grid(dens(i,:)==max(dens(i,:)));
        mean1(i)=sum(dens(i,:).*grid);
        std(i)=sqrt(sum(dens(i,:).*(grid-mean1(i)).^2)); 
        for j=1:length(quantiles)
            ind=find(cumdistribfunction<=quantiles(j),1,'last');
            if ~isempty(ind)
                quant(i,j)=grid(ind);
            else
                quant(i,j)=grid(1);
            end
        end
        if ~isempty(z)
            ind=find(grid<=z(i),1,'last');
            if isempty(ind)
                ind=1;
            end
            PIT(i)=cumdistribfunction(ind);
            LS(i)=-log(dens(i,ind));
        end
        
    end
    
    statistics.mean=mean1;
    statistics.std=std;
    statistics.modal=modal;
    statistics.median=median;
    statistics.quantiles=quant;
    statistics.PIT=PIT;
    statistics.LS=LS;
