function optimized=copulaEstimation(data,model,options)

% Performs maximum (composite) likelihood estimation for the model parameters based on
% a profile likelihood approach.
%
% Inputs:
%
% data........the data set as a structure array of the format data.x (x-coordinates), data.y
%             (y-coordinates), data.z (measured values)
%
% model.......contains the parameters of the spatial copula model (see
%             function initialize)
%
% options.....options structure array (see function testOptions)
%
% Outputs:
%
% optimized...contains the optimized parameters of the spatial copula model
%
% Hannes Kazianka, August 2011

    options=testOptions(options);

    if isempty(model.copula) || strcmp(model.copula.name,'Gaussian')
    	optimized=profilelikelihoodGaussian(model.margin,model.correlation,model.trend,model.anisotropy,data,options);
    elseif strcmp(model.copula.name,'Chi2')
        optimized=profilelikelihoodChi2(model.margin,model.correlation,model.trend,model.anisotropy,model.copula,data,options);
    elseif strcmp(model.copula.name,'Student')
    	optimized=profilelikelihoodStudent(model.margin,model.correlation,model.trend,model.anisotropy,model.copula,data,options);
    end

%%
%GAUSSIAN COPULA

function optimized= profilelikelihoodGaussian(margin,correlation,trend,anisotropy,data,options)
    
    if isempty(anisotropy.params)
        h=pdist([data.x data.y]);
        h=squareform(h);  
    elseif anisotropy.const==1
    	ab=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
        x=ab(1,:)';
        y=ab(2,:)';
        h=pdist([x y]);
        h=squareform(h);  
    end
    
    len=length(margin.params);
    lik=0;
    oldlik=1;
    
    if options.print>=1
        disp('Starting Profile Likelihood Approach');
    end
    
    while abs(lik-oldlik)>options.tol
        oldlik=lik;

        if ~isempty(anisotropy.params) && anisotropy.const==0
            ab=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
            x=ab(1,:)';
            y=ab(2,:)';
            h=pdist([x y]);
            h=squareform(h);   
        end
        
        cov=correlCovMat(h,correlation.params,correlation.model);
        R = chol(cov,'upper');
        Rinv=eye(size(cov)) / R;
        logSqrtDetSigma = sum(log(diag(R)));
        
        if ~isempty(margin.params) && ~isempty(trend.params) && (margin.const==0 || trend.const==0)
            [par lik]=fmincon(@optimfun1,[margin.params trend.params],[],[],[],[],[margin.lower trend.lower],[margin.upper trend.upper],[],options.optim,margin,trend,correlation,anisotropy,data,h,len,Rinv,logSqrtDetSigma,options);
            margin.params=par(1:len);
            trend.params=par((len+1):end);
        end
        
        if ~isempty(correlation.params) && correlation.const==0
            [correlation.params lik]=fmincon(@optimfun2,correlation.params,[],[],[],[],correlation.lower,correlation.upper,[],options.optim,margin,trend,correlation,anisotropy,data,h,options);
        end
 
        if ~isempty(anisotropy.params) && anisotropy.const==0
            [anisotropy.params lik]=fmincon(@optimfun3,anisotropy.params,[],[],[],[],anisotropy.lower,anisotropy.upper,[],options.optim,margin,trend,correlation,anisotropy,data,h,options);
        end  
        if options.print>=1
        	disp(['Profile Likelihood increase: ' num2str(abs(lik-oldlik))]);
            disp([num2str(lik) '   ' num2str(round([margin.params trend.params correlation.params anisotropy.params]*100)/100)]);
        end
    end
    
    if options.print>=1
        disp('Final result:');
        disp([num2str(lik) '   ' num2str(round([margin.params trend.params correlation.params anisotropy.params]*100)/100)]);
    end
    
    optimized.margin=margin;
    optimized.correlation=correlation;
    optimized.anisotropy=anisotropy;
    optimized.trend=trend;
    optimized.options=options;

    
function val =optimfun1(params,margin,trend,correlation,anisotropy,data,h,len,Rinv,logSqrtDetSigma,options)
        margin.params=params(1:len);
        trend.params=params((len+1):end);
        val=loglikelihoodGaussian(margin,trend,correlation,anisotropy,data,h,Rinv,logSqrtDetSigma,0,options);
        
function val =optimfun2(params,margin,trend,correlation,anisotropy,data,h,options)
        correlation.params=params;
        val=loglikelihoodGaussian(margin,trend,correlation,anisotropy,data,h,[],[],0,options);
                
function val =optimfun3(params,margin,trend,correlation,anisotropy,data,h,options)
        anisotropy.params=params;
        val=loglikelihoodGaussian(margin,trend,correlation,anisotropy,data,h,[],[],1,options);
        
    
function like=loglikelihoodGaussian(margin,trend,correlation,anisotropy,data,h,Rinv,logSqrtDetSigma,anisomode,options)

    if anisomode
    	xy=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
    	x=xy(1,:)';
    	y=xy(2,:)';
    end

    expected=trend.F*trend.params';
    if length(margin.params)==1
        z=feval([margin.name 'cdf'],data.z,expected,margin.params); 
        zz=feval([margin.name 'pdf'],data.z,expected,margin.params); 
    elseif length(margin.params)==2
        z=feval([margin.name 'cdf'],data.z,expected,margin.params(1),margin.params(2));
        zz=feval([margin.name 'pdf'],data.z,expected,margin.params(1),margin.params(2));
    end
    
    ok=sum(z~=1)==length(z);
    ok=ok*(sum(z~=0)==length(z));
    
    if ok==1
        z=norminv(z,0,1);
        if isempty(Rinv) || isempty(logSqrtDetSigma)
            if anisomode
                h=pdist([x y]);
                h=squareform(h);   
            end
            cov=correlCovMat(h,correlation.params,correlation.model);
            like=-mvnlogpdf_noconst(z',cov)-0.5*(z'*z)-sum(log(zz));
        else
            xRinv = z'*Rinv;
            quadform = sum(xRinv.^2, 2);
            like=0.5*quadform+logSqrtDetSigma-0.5*(z'*z)-sum(log(zz));
        end

        if isnan(like)
            like=10^100;%a large value. This is to continue the optimization procedure if inappropriate values for the parameters are chosen by the fmincon-algorithm
        end
    else
        like=10^100;
    end    
    if options.print>=2
        disp([num2str(like) ' ' num2str(round([margin.params trend.params correlation.params anisotropy.params]*1000)/1000)])     
    end
   
%%
%CHI^2-COPULA
 
 
function optimized= profilelikelihoodChi2(margin,correlation,trend,anisotropy,copula,data,options)
    
    if isempty(anisotropy.params)
        h=pdist([data.x data.y]);
        h=squareform(h);  
    elseif anisotropy.const==1
    	ab=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
        x=ab(1,:)';
        y=ab(2,:)';
        h=pdist([x y]);
        h=squareform(h);  
    end
    
    
    len=length(margin.params);
    lik=0;
    oldlik=1;
    
    
    while abs(lik-oldlik)>options.tol
    	oldlik=lik;

        if ~isempty(anisotropy.params) && anisotropy.const==0
            ab=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
            x=ab(1,:)';
            y=ab(2,:)';
            h=pdist([x y]);
            h=squareform(h);   
        end

        if ~isempty(margin.params) && ~isempty(trend.params) && (margin.const==0 || trend.const==0)
            [par lik]=fmincon(@optimfun4,[margin.params trend.params],[],[],[],[],[margin.lower trend.lower],[margin.upper trend.upper],[],options.optim,margin,trend,correlation,anisotropy,copula,data,h,len,options);
            margin.params=par(1:len);
            trend.params=par((len+1):end);
        end
        
        if ~isempty(copula.params) && copula.const==0
            [copula.params lik]=fmincon(@optimfun7,copula.params,[],[],[],[],copula.lower,copula.upper,[],options.optim,margin,trend,correlation,anisotropy,copula,data,h,options);
        end
        
        if ~isempty(correlation.params) && correlation.const==0
           [correlation.params lik]=fmincon(@optimfun5,correlation.params,[],[],[],[],correlation.lower,correlation.upper,[],options.optim,margin,trend,correlation,anisotropy,copula,data,h,options);
        end
        
        if ~isempty(anisotropy.params) && anisotropy.const==0
           [anisotropy.params lik]=fmincon(@optimfun6,anisotropy.params,[],[],[],[],anisotropy.lower,anisotropy.upper,[],options.optim,margin,trend,correlation,anisotropy,copula,data,h,options);
        end     

        if options.print>=1
        	disp(['Profile Likelihood increase: ' num2str(abs(lik-oldlik))]);
            disp([num2str(lik) '   ' num2str(round([margin.params trend.params correlation.params anisotropy.params copula.params]*1000)/1000)])
        end
    end

    if options.print>=1
        disp('Final result:');
        disp([num2str(lik) '   ' num2str(round([margin.params trend.params correlation.params anisotropy.params copula.params]*100)/100)])
    end
    
    optimized.margin=margin;
    optimized.correlation=correlation;
    optimized.anisotropy=anisotropy;
    optimized.trend=trend;
    optimized.copula=copula;
    optimized.options=options;

    
function val =optimfun4(params,margin,trend,correlation,anisotropy,copula,data,h,len,options)
 	margin.params=params(1:len);
	trend.params=params((len+1):end);
	val=loglikelihoodChi2(margin,trend,correlation,anisotropy,copula,data,h,0,options);
        
function val =optimfun5(params,margin,trend,correlation,anisotropy,copula,data,h,options)
    correlation.params=params;
    val=loglikelihoodChi2(margin,trend,correlation,anisotropy,copula,data,h,0,options);
                
function val =optimfun6(params,margin,trend,correlation,anisotropy,copula,data,h,options)
    anisotropy.params=params;
    val=loglikelihoodChi2(margin,trend,correlation,anisotropy,copula,data,h,1,options);
    
function val =optimfun7(params,margin,trend,correlation,anisotropy,copula,data,h,options)
    copula.params=params;
    val=loglikelihoodChi2(margin,trend,correlation,anisotropy,copula,data,h,0,options);
        
        
    
function like = loglikelihoodChi2(margin,trend,correlation,anisotropy,copula,data,h,anisomode,options)
    
    if anisomode
            xy=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
            x=xy(1,:)';
            y=xy(2,:)';
    end

    expected=trend.F*trend.params';
    if length(margin.params)==1
        z=feval([margin.name 'cdf'],data.z,expected,margin.params); 
        zz=feval([margin.name 'pdf'],data.z,expected,margin.params); 
    elseif length(margin.params)==2
        z=feval([margin.name 'cdf'],data.z,expected,margin.params(1),margin.params(2));
        zz=feval([margin.name 'pdf'],data.z,expected,margin.params(1),margin.params(2));
    end
    
    n=length(data.z);
    ok=sum(z~=1)==n;
    ok=ok*(sum(z~=0)==n);
    lambda=copula.params;
    
    if ok==1
        if anisomode
            h=pdist([x y]);
            h=squareform(h);   
        end

        if lambda<0
            like=10^100;
        else
            z=ncx2inv(z,1,lambda); 
            x=log(ncx2pdf(z,1,lambda));
            z=sqrt(z);
            logz=log(z);
            lambda=sqrt(lambda);
            %m=[lambda,lambda];      
            
            zz=log(zz);
            like=0;
            cov=correlCovMat(h,correlation.params,correlation.model);
            for i=1:n
                for j=(i+1):n
                    %like=like-(log(sum(mvnpdf([z(i),z(j);-z(i),z(j);z(i),-z(j);-z(i),-z(j)],m,cov([i,j],[i,j]))))-logz(i)-logz(j)-x(i)-x(j))-zz(i)-zz(j);
                    like=like-(log(sum(exp(mvnlogpdf_noconst([z(i),z(j);-z(i),z(j);z(i),-z(j);-z(i),-z(j)]-lambda,cov([i,j],[i,j]))))/(2*pi))-logz(i)-logz(j)-x(i)-x(j))-zz(i)-zz(j);
                end
            end
            if isnan(like) || isinf(like)
                like=10^100;
            end
        end
    else
        like=10^100;
    end
    
    if options.print>=2
        disp([num2str(like) ' ' num2str(round([margin.params trend.params correlation.params anisotropy.params copula.params]*1000)/1000)])
    end
    
%%   
%Student-t copula
function optimized= profilelikelihoodStudent(margin,correlation,trend,anisotropy,copula,data,options)
    
    if isempty(anisotropy.params)
        h=pdist([data.x data.y]);
        h=squareform(h);  
    elseif anisotropy.const==1
    	ab=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
        x=ab(1,:)';
        y=ab(2,:)';
        h=pdist([x y]);
        h=squareform(h); 
    end
        
    len=length(margin.params);
    lik=0;
    oldlik=1;

    while abs(lik-oldlik)>options.tol
        oldlik=lik;
        
        if ~isempty(anisotropy.params) && anisotropy.const==0
            ab=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
            x=ab(1,:)';
            y=ab(2,:)';
            h=pdist([x y]);
            h=squareform(h);   
        end
        
        cov=correlCovMat(h,correlation.params,correlation.model);
        R = chol(cov,'upper');
        Rinv=eye(size(cov)) / R;
        logSqrtDetSigma = sum(log(diag(R)));
        
        if ~isempty(margin.params) && ~isempty(trend.params) && (margin.const==0 || trend.const==0)
            [par lik]=fmincon(@optimfun8,[margin.params trend.params],[],[],[],[],[margin.lower trend.lower],[margin.upper trend.upper],[],options.optim,margin,trend,correlation,anisotropy,copula,data,h,len,Rinv,logSqrtDetSigma,options);
            margin.params=par(1:len);
            trend.params=par((len+1):end);
        end
        
        if ~isempty(copula.params) && copula.const==0
            [copula.params lik]=fmincon(@optimfun9,copula.params,[],[],[],[],copula.lower,copula.upper,[],options.optim,margin,trend,correlation,anisotropy,copula,data,h,Rinv,logSqrtDetSigma,options);
        end  
        
        if ~isempty(correlation.params) && correlation.const==0
            [correlation.params lik]=fmincon(@optimfun10,correlation.params,[],[],[],[],correlation.lower,correlation.upper,[],options.optim,margin,trend,correlation,anisotropy,copula,data,h,options);  
        end
        
        if ~isempty(anisotropy.params) && anisotropy.const==0
            [anisotropy.params lik]=fmincon(@optimfun11,anisotropy.params,[],[],[],[],anisotropy.lower,anisotropy.upper,[],options.optim,margin,trend,correlation,anisotropy,copula,data,h,options);  
        end

        if options.print>=1
        	disp(['Profile Likelihood increase: ' num2str(abs(lik-oldlik))]);
            disp([num2str(lik) '   ' num2str(round([margin.params trend.params correlation.params anisotropy.params copula.params]*1000)/1000)])
        end
    end
    
    if options.print>=1
        disp('Final result:');
        disp([num2str(lik) '   ' num2str(round([margin.params trend.params correlation.params anisotropy.params copula.params]*100)/100)])
    end
    
    optimized.margin=margin;
    optimized.correlation=correlation;
    optimized.anisotropy=anisotropy;
    optimized.trend=trend;
    optimized.copula=copula;
    optimized.options=options;
    
    
function val =optimfun8(params,margin,trend,correlation,anisotropy,copula,data,h,len,Rinv,logSqrtDetSigma,options)
        margin.params=params(1:len);
        trend.params=params((len+1):end);
        val=loglikelihoodStudent(margin,trend,correlation,anisotropy,copula,data,h,Rinv,logSqrtDetSigma,0,options);

function val =optimfun9(params,margin,trend,correlation,anisotropy,copula,data,h,Rinv,logSqrtDetSigma,options)
        copula.params=params;
        val=loglikelihoodStudent(margin,trend,correlation,anisotropy,copula,data,h,Rinv,logSqrtDetSigma,0,options);
        

function val =optimfun10(params,margin,trend,correlation,anisotropy,copula,data,h,options)
        correlation.params=params;
        val=loglikelihoodStudent(margin,trend,correlation,anisotropy,copula,data,h,[],[],0,options);
        
function val =optimfun11(params,margin,trend,correlation,anisotropy,copula,data,h,options)
        anisotropy.params=params;
        val=loglikelihoodStudent(margin,trend,correlation,anisotropy,copula,data,h,[],[],1,options);
    
function like = loglikelihoodStudent(margin,trend,correlation,anisotropy,copula,data,h,Rinv,logSqrtDetSigma,anisomode,options)

    if anisomode
            xy=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
            x=xy(1,:)';
            y=xy(2,:)';
    end
  
    
    expected=trend.F*trend.params';
    if length(margin.params)==1
        z=feval([margin.name 'cdf'],data.z,expected,margin.params); 
        zz=feval([margin.name 'pdf'],data.z,expected,margin.params); 
    elseif length(margin.params)==2
        z=feval([margin.name 'cdf'],data.z,expected,margin.params(1),margin.params(2));
        zz=feval([margin.name 'pdf'],data.z,expected,margin.params(1),margin.params(2));
    end
    
    n=length(z);    
    ok=sum(z~=1)==n;
    ok=ok*(sum(z~=0)==n);
    nu=copula.params;
    
    if ok==1
        z=tinv(z,nu);
        zzz=tpdf(z,nu);
        if isempty(Rinv) || isempty(logSqrtDetSigma)
            if anisomode
                h=pdist([x y]);
                h=squareform(h);   
            end
            cov=correlCovMat(h,correlation.params,correlation.model);
            R = chol(cov,'upper');
            xRinv = z' / R;
            logSqrtDetSigma = sum(log(diag(R)));
            quadform = sum(xRinv.^2, 2);
            like=gammaln(nu/2)-gammaln((nu+n)/2)+n/2*log(nu)+0.5*(nu+n)*log(1+1/nu*quadform)+logSqrtDetSigma-sum(log(zz))+sum(log(zzz));
        else
            xRinv = z'*Rinv;
            quadform = sum(xRinv.^2, 2);
            like=gammaln(nu/2)-gammaln((nu+n)/2)+n/2*log(nu)+0.5*(nu+n)*log(1+1/nu*quadform)+logSqrtDetSigma-sum(log(zz))+sum(log(zzz));
        end

        if isnan(like)
            like=10^100;
        end
    else
        like=10^100;
    end        

    if options.print>=2
        disp([num2str(like) ' ' num2str(round([margin.params trend.params correlation.params anisotropy.params copula.params]*1000)/1000)])
    end 
   