function optimized=copulaEstimationDiscrete2(data,model,options,method)

% Performs GQT or composite likelihood estimation for the model parameters based on
% a profile likelihood approach in case of discrete data. Permissible
% margin.names are 'poiss', 'bino' and 'geo'.
%
% Inputs:
%
% data........the data set as a structure array of the format data.x (x-coordinates), data.y
%             (y-coordinates), data.z (measured values)
%
% model.......contains the parameters of the spatial copula model (see
%             function initialize)
%
% options.....options structure array (see function testOptions)
%
% method......string indicating whether the GQT method (method='GQT') should be used
%             to estimate the parameters or not. In the latter case, the
%             composite likelihood approach is used.
%
% Outputs:
%
% optimized...contains the optimized parameters of the spatial copula model
%
% Hannes Kazianka, August 2012
    if nargin<4
       method='CL'; 
    end
    options=testOptions(options);
    
    if isempty(model.copula) || strcmp(model.copula.name,'Gaussian')
        optimized=profilelikelihoodGaussian(model.margin,model.correlation,model.trend,model.anisotropy,data,options,method);
    elseif strcmp(model.copula.name,'Chi2') && strcmp(method,'GQT')
        optimized=profilelikelihoodChi2(model.margin,model.correlation,model.trend,model.anisotropy,model.copula,data,options,method);
    elseif strcmp(model.copula.name,'Student')
        optimized=profilelikelihoodStudent(model.margin,model.correlation,model.trend,model.anisotropy,model.copula,data,options,method);
    end
%%
%GAUSSIAN COPULA

function optimized= profilelikelihoodGaussian(margin,correlation,trend,anisotropy,data,options,method)
    
    if isempty(anisotropy.params)
        h=pdist([data.x data.y]);
        h=squareform(h);  
    elseif anisotropy.const==1
    	ab=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
        x=ab(1,:)';
        y=ab(2,:)';
        h=pdist([x y]);
        h=squareform(h);  
    end
    
    lik=0;
    oldlik=1;

    if options.print>=1
        disp('Starting Profile Likelihood Approach');
    end

    
    while abs(lik-oldlik)>options.tol
        oldlik=lik;

        if ~isempty(anisotropy.params) && anisotropy.const==0
            ab=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
            a=ab(1,:)';
            b=ab(2,:)';
            h=pdist([a b]);
            h=squareform(h);   
        end
        
        cov=correlCovMat(h,correlation.params,correlation.model);
        if strcmp(method,'GQT')
            R = chol(cov,'upper');
            Rinv=eye(size(cov)) / R;
            logSqrtDetSigma = sum(log(diag(R)));
        else
            Rinv=[];
            logSqrtDetSigma=[];
        end        
        
        if ~isempty(trend.params) &&  trend.const==0
            [trend.params lik]=fmincon(@optimfun1,trend.params,[],[],[],[],trend.lower,trend.upper,[],options,margin,trend,correlation,anisotropy,data,h,Rinv,logSqrtDetSigma,cov,options,method);
        end    

        if ~isempty(correlation.params) && correlation.const==0
            [correlation.params lik]=fmincon(@optimfun2,correlation.params,[],[],[],[],correlation.lower,correlation.upper,[],options,margin,trend,correlation,anisotropy,data,h,options,method);
        end
 
        if ~isempty(anisotropy.params) && anisotropy.const==0
            [anisotropy.params lik]=fmincon(@optimfun3,anisotropy.params,[],[],[],[],anisotropy.lower,anisotropy.upper,[],options,margin,trend,correlation,anisotropy,data,h,options,method);
        end  
      
        if options.print>=1
        	disp(['Profile Likelihood increase: ' num2str(abs(lik-oldlik))]);
            disp([num2str(lik) '   ' num2str(round([trend.params correlation.params anisotropy.params]*100)/100)]);
        end
    end
    
    if options.print>=1
        disp('Final result:');
        disp([num2str(lik) '   ' num2str(round([trend.params correlation.params anisotropy.params]*100)/100)]);
    end
    
    optimized.margin=margin;
    optimized.correlation=correlation;
    optimized.anisotropy=anisotropy;
    optimized.trend=trend;
    optimized.options=options;

    
function val =optimfun1(params,margin,trend,correlation,anisotropy,data,h,Rinv,logSqrtDetSigma,cov,options,method)
        trend.params=params;
        val=logoptimfunGaussian(margin,trend,correlation,anisotropy,data,h,Rinv,logSqrtDetSigma,cov,0,options,method);
        
function val =optimfun2(params,margin,trend,correlation,anisotropy,data,h,options,method)
        correlation.params=params;
        val=logoptimfunGaussian(margin,trend,correlation,anisotropy,data,h,[],[],[],0,options,method);
                
function val =optimfun3(params,margin,trend,correlation,anisotropy,data,h,options,method)
        anisotropy.params=params;
        val=logoptimfunGaussian(margin,trend,correlation,anisotropy,data,h,[],[],[],1,options,method);


function like=logoptimfunGaussian(margin,trend,correlation,anisotropy,data,h,Rinv,logSqrtDetSigma,cov,anisomode,options,method)

    if anisomode
            xy=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
            x=xy(1,:)';
            y=xy(2,:)';
    end

    expected=trend.F*trend.params';
    if strcmp(margin.name,'poiss')
        expected=exp(expected);
        zz=poisspdf(data.z,expected); 
        b=poisscdf(data.z-1,expected); 
    elseif strcmp(margin.name,'bino')
        expected=exp(expected)./(1+exp(expected));
        zz=binopdf(data.z,margin.paramsfix,expected); %paramsfix=verschiedene n_i
        b=binocdf(data.z-1,margin.paramsfix,expected); 
    elseif strcmp(margin.name,'geo')
        expected=exp(expected)./(1+exp(expected));
        zz=geopdf(data.z,expected);
        b=geocdf(data.z-1,expected);
    end

    ok=sum(isinf(b))==0;
    if ok
        if anisomode
            h=pdist([x y]);
            h=squareform(h);   
        end

        if strcmp(method,'GQT')                
            z=b+zz/2;
            z=norminv(z);
            if isempty(Rinv) || isempty(logSqrtDetSigma)
                cov=correlCovMat(h,correlation.params,correlation.model);
                like=-mvnlogpdf_noconst(z',cov)-0.5*(z'*z)-sum(log(zz)); 
            else
                xRinv = z'*Rinv;
                quadform = sum(xRinv.^2, 2);
                like=0.5*quadform+logSqrtDetSigma-0.5*(z'*z)-sum(log(zz));
            end
        else
            a=norminv(b+zz,0,1);
            b=norminv(b,0,1); 
            like=0;
            if isempty(cov)
                cov=correlCovMat(h,correlation.params,correlation.model);
            end
            try
                for i=1:length(a)
                    for j=(i+1):length(a)
                        temp=mvncdf([a(i),a(j);a(i),b(j);b(i),a(j);b(i),b(j)],[0 0],[1 cov(i,j);cov(i,j) 1]);
                        like=like-log(temp(1)-temp(2)-temp(3)+temp(4));
                    end
                end     
            catch
                like=Inf;
            end
        end
        if isnan(like) || isinf(like)
            like=10^100;%a large value. This is to continue the optimization procedure if inappropriate values for the parameters are chosen by the fmincon-algorithm
        end
    else 
        like=10^100;
    end
    if options.print>=2
        disp([num2str(like) ' ' num2str(round([trend.params correlation.params anisotropy.params]*1000)/1000)])
    end
    
%%
%Student-COPULA

function optimized= profilelikelihoodStudent(margin,correlation,trend,anisotropy,copula,data,options,method)
    
    if isempty(anisotropy.params)
        h=pdist([data.x data.y]);
        h=squareform(h);  
    elseif anisotropy.const==1
    	ab=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
        x=ab(1,:)';
        y=ab(2,:)';
        h=pdist([x y]);
        h=squareform(h);  
    end
    
    lik=0;
    oldlik=1;

    if options.print>=1
        disp('Starting Profile Likelihood Approach');
    end

    
    while abs(lik-oldlik)>options.tol
        oldlik=lik;

        if ~isempty(anisotropy.params) && anisotropy.const==0
            ab=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
            a=ab(1,:)';
            b=ab(2,:)';
            h=pdist([a b]);
            h=squareform(h);   
        end
        
        cov=correlCovMat(h,correlation.params,correlation.model);
        if strcmp(method,'GQT')
            R = chol(cov,'upper');
            Rinv=eye(size(cov)) / R;
            logSqrtDetSigma = sum(log(diag(R)));
        else
            Rinv=[];
            logSqrtDetSigma=[];
        end        
        
        if ~isempty(trend.params) &&  trend.const==0
            [trend.params lik]=fmincon(@optimfun8,trend.params,[],[],[],[],trend.lower,trend.upper,[],options,margin,trend,correlation,anisotropy,copula,data,h,Rinv,logSqrtDetSigma,cov,options,method);
        end
        
        if ~isempty(copula.params) &&  copula.const==0
            [copula.params lik]=fmincon(@optimfun9,copula.params,[],[],[],[],trend.lower,trend.upper,[],options,margin,trend,correlation,anisotropy,copula,data,h,Rinv,logSqrtDetSigma,cov,options,method);
        end
        
        if ~isempty(correlation.params) && correlation.const==0
            [correlation.params lik]=fmincon(@optimfun10,correlation.params,[],[],[],[],correlation.lower,correlation.upper,[],options,margin,trend,correlation,anisotropy,copula,data,h,options,method);
        end
 
        if ~isempty(anisotropy.params) && anisotropy.const==0
            [anisotropy.params lik]=fmincon(@optimfun11,anisotropy.params,[],[],[],[],anisotropy.lower,anisotropy.upper,[],options,margin,trend,correlation,anisotropy,copula,data,h,options,method);
        end  
      
        if options.print>=1
        	disp(['Profile Likelihood increase: ' num2str(abs(lik-oldlik))]);
            disp([num2str(lik) '   ' num2str(round([trend.params correlation.params anisotropy.params copula.params]*100)/100)]);
        end
    end
    
    if options.print>=1
        disp('Final result:');
        disp([num2str(lik) '   ' num2str(round([trend.params correlation.params anisotropy.params copula.params]*100)/100)]);
    end
    
    optimized.margin=margin;
    optimized.correlation=correlation;
    optimized.anisotropy=anisotropy;
    optimized.trend=trend;
    optimized.copla=copula;
    optimized.options=options;

    
function val =optimfun8(params,margin,trend,correlation,anisotropy,copula,data,h,Rinv,logSqrtDetSigma,cov,options,method)
        trend.params=params;
        val=logoptimfunStudent(margin,trend,correlation,anisotropy,copula,data,h,Rinv,logSqrtDetSigma,cov,0,options,method);
        
function val =optimfun9(params,margin,trend,correlation,anisotropy,copula,data,h,Rinv,logSqrtDetSigma,cov,options,method)
        copula.params=params;
        val=logoptimfunStudent(margin,trend,correlation,anisotropy,copula,data,h,Rinv,logSqrtDetSigma,cov,0,options,method);
        
function val =optimfun10(params,margin,trend,correlation,anisotropy,copula,data,h,options,method)
        correlation.params=params;
        val=logoptimfunStudent(margin,trend,correlation,anisotropy,copula,data,h,[],[],[],0,options,method);
                
function val =optimfun11(params,margin,trend,correlation,anisotropy,copula,data,h,options,method)
        anisotropy.params=params;
        val=logoptimfunStudent(margin,trend,correlation,anisotropy,copula,data,h,[],[],[],1,options,method);


function like=logoptimfunStudent(margin,trend,correlation,anisotropy,copula,data,h,Rinv,logSqrtDetSigma,cov,anisomode,options,method)

    if anisomode
            xy=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
            x=xy(1,:)';
            y=xy(2,:)';
    end

    expected=trend.F*trend.params';
    if strcmp(margin.name,'poiss')
        expected=exp(expected);
        zz=poisspdf(data.z,expected); 
        b=poisscdf(data.z-1,expected); 
    elseif strcmp(margin.name,'bino')
        expected=exp(expected)./(1+exp(expected));
        zz=binopdf(data.z,margin.paramsfix,expected); %paramsfix=verschiedene n_i
        b=binocdf(data.z-1,margin.paramsfix,expected); 
    elseif strcmp(margin.name,'geo')
        expected=exp(expected)./(1+exp(expected));
        zz=geopdf(data.z,expected);
        b=geocdf(data.z-1,expected);
    end

    ok=sum(isinf(b))==0;
    if ok
        if anisomode
            h=pdist([x y]);
            h=squareform(h);   
        end
        n=length(zz);  
        nu=copula.params;
        if strcmp(method,'GQT')                
            z=b+zz/2;
            z=tinv(z,nu);
            zzz=tpdf(z,nu);
            if isempty(Rinv) || isempty(logSqrtDetSigma)
                cov=correlCovMat(h,correlation.params,correlation.model);
                R = chol(cov,'upper');
                xRinv = z' / R;
                logSqrtDetSigma = sum(log(diag(R)));
                quadform = sum(xRinv.^2, 2);
                like=gammaln(nu/2)-gammaln((nu+n)/2)+n/2*log(nu)+0.5*(nu+n)*log(1+1/nu*quadform)+logSqrtDetSigma-sum(log(zz))+sum(log(zzz));
            else
                xRinv = z'*Rinv;
                quadform = sum(xRinv.^2, 2);
                like=gammaln(nu/2)-gammaln((nu+n)/2)+n/2*log(nu)+0.5*(nu+n)*log(1+1/nu*quadform)+logSqrtDetSigma-sum(log(zz))+sum(log(zzz));
            end
        else
            a=tinv(b+zz,nu);
            b=tinv(b,nu); 
            like=0;
            if isempty(cov)
                cov=correlCovMat(h,correlation.params,correlation.model);
            end
            for i=1:length(a)
                for j=(i+1):length(a)
                    temp=mvtcdf([a(i),a(j);a(i),b(j);b(i),a(j);b(i),b(j)],[1 cov(i,j);cov(i,j) 1],nu);
                    like=like-log(temp(1)-temp(2)-temp(3)+temp(4));
                end
            end     
        end
        if isnan(like) || isinf(like)
            like=10^100;%a large value. This is to continue the optimization procedure if inappropriate values for the parameters are chosen by the fmincon-algorithm
        end
    else 
        like=10^100;
    end
    if options.print>=2
        disp([num2str(like) ' ' num2str(round([trend.params correlation.params anisotropy.params copula.params]*1000)/1000)])
    end

%%
%Chi2-COPULA

function optimized= profilelikelihoodChi2(margin,correlation,trend,anisotropy,copula,data,options,method)
    
    if isempty(anisotropy.params)
        h=pdist([data.x data.y]);
        h=squareform(h);  
    elseif anisotropy.const==1
    	ab=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
        x=ab(1,:)';
        y=ab(2,:)';
        h=pdist([x y]);
        h=squareform(h);  
    end
    
    lik=0;
    oldlik=1;

    if options.print>=1
        disp('Starting Profile Likelihood Approach');
    end

    
    while abs(lik-oldlik)>options.tol
        oldlik=lik;

        if ~isempty(anisotropy.params) && anisotropy.const==0
            ab=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
            a=ab(1,:)';
            b=ab(2,:)';
            h=pdist([a b]);
            h=squareform(h);   
        end
        
        cov=correlCovMat(h,correlation.params,correlation.model);
        
        if ~isempty(trend.params) &&  trend.const==0
            [trend.params lik]=fmincon(@optimfun4,trend.params,[],[],[],[],trend.lower,trend.upper,[],options,margin,trend,correlation,anisotropy,copula,data,h,cov,options,method);
        end
        
        if ~isempty(copula.params) &&  copula.const==0
            [copula.params lik]=fmincon(@optimfun5,copula.params,[],[],[],[],trend.lower,trend.upper,[],options,margin,trend,correlation,anisotropy,copula,data,h,cov,options,method);
        end
        
        if ~isempty(correlation.params) && correlation.const==0
            [correlation.params lik]=fmincon(@optimfun6,correlation.params,[],[],[],[],correlation.lower,correlation.upper,[],options,margin,trend,correlation,anisotropy,copula,data,h,options,method);
        end
 
        if ~isempty(anisotropy.params) && anisotropy.const==0
            [anisotropy.params lik]=fmincon(@optimfun7,anisotropy.params,[],[],[],[],anisotropy.lower,anisotropy.upper,[],options,margin,trend,correlation,anisotropy,copula,data,h,options,method);
        end  
      
        if options.print>=1
        	disp(['Profile Likelihood increase: ' num2str(abs(lik-oldlik))]);
            disp([num2str(lik) '   ' num2str(round([trend.params correlation.params anisotropy.params copula.params]*100)/100)]);
        end
    end
    
    if options.print>=1
        disp('Final result:');
        disp([num2str(lik) '   ' num2str(round([trend.params correlation.params anisotropy.params copula.params]*100)/100)]);
    end
    
    optimized.margin=margin;
    optimized.correlation=correlation;
    optimized.anisotropy=anisotropy;
    optimized.trend=trend;
    optimized.copla=copula;
    optimized.options=options;

    
function val =optimfun4(params,margin,trend,correlation,anisotropy,copula,data,h,cov,options,method)
        trend.params=params;
        val=logoptimfunChi2(margin,trend,correlation,anisotropy,copula,data,h,cov,0,options,method);
        
function val =optimfun5(params,margin,trend,correlation,anisotropy,copula,data,h,cov,options,method)
        copula.params=params;
        val=logoptimfunChi2(margin,trend,correlation,anisotropy,copula,data,h,cov,0,options,method);
        
function val =optimfun6(params,margin,trend,correlation,anisotropy,copula,data,h,options,method)
        correlation.params=params;
        val=logoptimfunChi2(margin,trend,correlation,anisotropy,copula,data,h,[],0,options,method);
                
function val =optimfun7(params,margin,trend,correlation,anisotropy,copula,data,h,options,method)
        anisotropy.params=params;
        val=logoptimfunChi2(margin,trend,correlation,anisotropy,copula,data,h,[],1,options,method);


function like=logoptimfunChi2(margin,trend,correlation,anisotropy,copula,data,h,cov,anisomode,options,method)

    if anisomode
            xy=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
            x=xy(1,:)';
            y=xy(2,:)';
    end

    expected=trend.F*trend.params';
    if strcmp(margin.name,'poiss')
        expected=exp(expected);
        zz=poisspdf(data.z,expected); 
        b=poisscdf(data.z-1,expected); 
    elseif strcmp(margin.name,'bino')
        expected=exp(expected)./(1+exp(expected));
        zz=binopdf(data.z,margin.paramsfix,expected); %paramsfix=verschiedene n_i
        b=binocdf(data.z-1,margin.paramsfix,expected); 
    elseif strcmp(margin.name,'geo')
        expected=exp(expected)./(1+exp(expected));
        zz=geopdf(data.z,expected);
        b=geocdf(data.z-1,expected);
    end

    ok=sum(isinf(b))==0;
    if ok && strcmp(method,'GQT')
        if anisomode
            h=pdist([x y]);
            h=squareform(h);   
        end
        n=length(zz);  
        lambda=copula.params;
        z=b+zz/2;
        z=ncx2inv(z,1,lambda); 
        x=log(ncx2pdf(z,1,lambda));
        z=sqrt(z);
        logz=log(z);
        lambda=sqrt(lambda);
        zz=log(zz);
        like=0;
        if isempty(cov)
            cov=correlCovMat(h,correlation.params,correlation.model);
        end
        for i=1:n
            for j=(i+1):n
                like=like-(log(sum(exp(mvnlogpdf_noconst([z(i),z(j);-z(i),z(j);z(i),-z(j);-z(i),-z(j)]-lambda,cov([i,j],[i,j]))))/(2*pi))-logz(i)-logz(j)-x(i)-x(j))-zz(i)-zz(j);
            end
        end

        if isnan(like) || isinf(like)
            like=10^100;%a large value. This is to continue the optimization procedure if inappropriate values for the parameters are chosen by the fmincon-algorithm
        end
    else 
        like=10^100;
    end
    if options.print>=2
        disp([num2str(like) ' ' num2str(round([trend.params correlation.params anisotropy.params copula.params]*1000)/1000)])
    end