function output = copulaIK(locations,data,model,thres)

% Performs copula-based indicator kriging
%
% Inputs:
%
% locations...structure array representing the x-coordinates (locations.x),
%             y-coordinates (locations.y) and design matrix (locations.F)
%             of locations where prediction is desired
%
% data........the data set as a structure array of the format data.x (x-coordinates), data.y
%             (y-coordinates), data.z (measured values)
%
% model.......contains the parameters of the spatial copula model (see
%             function initialize)
%
% thres.......vector of thresholds, for which the indicator kriging
%             (prediction of P(Z(x)<=thres)) is performed
%
% Outputs:
%
% output......structure array consisting of the indicator kriging output
%             and the thresholds
%             output.pred......matrix (number of locations x number of
%                              thresholds) consisting of the indicator 
%                              kriging results 
%            
%             output.thres.....the thresholds
%  
% Hannes Kazianka, August 2011

    margin=model.margin;
    trend=model.trend;
    correlation=model.correlation;
    anisotropy=model.anisotropy;
    copula=model.copula;
    
    origlocations.x=locations.x;
    origlocations.y=locations.y;
    if ~isempty(anisotropy)
        xy=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
        data.x=xy(1,:)';
        data.y=xy(2,:)';
        xy=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[locations.x';locations.y'];
        locations.x=xy(1,:)';
        locations.y=xy(2,:)';
    end
    n=length(data.x);
    k=length(thres);

    expected=trend.F*trend.params';
    thresnew=zeros(n,k);
    for i=1:k
        if length(margin.params)==1
            thresnew(:,i)=feval([margin.name 'cdf'],thres(i),expected,margin.params);
        elseif length(margin.params)==2
            thresnew(:,i)=feval([margin.name 'cdf'],thres(i),expected,margin.params(1),margin.params(2));
        end
    end

    h=pdist([[data.x;locations.x] [data.y;locations.y]]);
    h=squareform(h);


    Z=zeros(1,n*k);
    for i=1:n
        Z(((i-1)*k+1):(i*k))=data.z(i)<=thres;
    end

    F=zeros(n*k,n*k);

    for istrich=1:n
        for i=1:n
            F(((1:k)-1)*n+istrich,(i-1)*k+(1:k))=reshape(covariance(h(istrich,i),kron(thresnew(i,:),ones(1,k))',kron(ones(1,k),thresnew(istrich,:))',correlation.params,correlation.model,copula),k,k);
        end
    end

    app=zeros(k,n*k);
    for i=1:n
        app(:,((i-1)*k+1):(i*k))=eye(k);
    end

    F=[F;app];

    app=zeros((n+1)*k,k);
    for i=1:k
        app(((i-1)*n+1):(i*n),i)=-ones(n,1);
    end
    F=[F app];

    Finv=inv(F);
    Zhat=zeros(length(locations.x),length(thres));


    expected=locations.F*trend.params';
    thresx0=zeros(length(expected),k);
    for i=1:k
        if length(margin.params)==1
            thresx0(:,i)=feval([margin.name 'cdf'],thres(i),expected,margin.params);
        elseif length(margin.params)==2
            thresx0(:,i)=feval([margin.name 'cdf'],thres(i),expected,margin.params(1),margin.params(2));
        end
    end



    for i=1:length(locations.x)
        if ~isnan(locations.x(i))
            for j=1:length(thres)
                rightside=zeros(n*k,1);
                for jstrich=1:k
                    for istrich=1:n
                        rightside(istrich+(jstrich-1)*n)=covariance(h(istrich,n+i),thresx0(i,j),thresnew(istrich,jstrich),correlation.params,correlation.model);
                    end
                end
                app=zeros(k,1);
                app(j)=1;
                rightside=[rightside;app];
                lambda=Finv*rightside;
                lambda=lambda(1:n*k);
                Zhat(i,j)=Z*lambda;
            end
        end
    end

    output.pred=Zhat;
    output.thres=thres;



function val = covariance(h,q1,q2,param,model,copula)
    if h~=0
        if isempty(copula) || strcmp(copula.name,'Gaussian')
            val=gausscop([q1,q2],h,param,model)-q1.*q2;
        else
            switch copula.name
                case 'Chi2' 
                    val=chi2cop([q1,q2],h,param,model,copula.params)-q1.*q2;
                case 'Student'
                    val=studentcop([q1,q2],h,param,model,copula.params)-q1.*q2;
            end
        end
    else
        val=min([q1;q2])-q1.*q2;
    end
   

function val = gausscop(x,h,delta,model)

    cov=correlCovMat(h,delta,model);
    sigma=[1 cov;cov 1];
    x=norminv(x,0,1);
    val=mvncdf(x,[0 0],sigma);
    
function val =studentcop(x,h,delta,model,nu)
    cov=correlCovMat(h,delta,model);
    sigma=[1 cov;cov 1];
    x=tinv(x,nu);
    val=mvtcdf(x,sigma,nu);
        
function val =chi2cop(x,h,delta,model,lambda)
    cov=correlCovMat(h,delta,model);
    sigma=[1 cov;cov 1];
    x=sqrt(ncx2inv(x,1,lambda));
    m=[sqrt(lambda) sqrt(lambda)];
    val=mvncdf(x,m,sigma)-mvncdf([x(:,1) -x(:,2)],m,sigma)-mvncdf([-x(:,1) x(:,2)],m,sigma)+mvncdf(-x,m,sigma);