function prediction = copulaPredict(locations,data,model,search,calc,options)

% Performs copula-based spatial prediction at given locations
%
% Inputs:
%
% locations...structure array representing the x-coordinates (locations.x),
%             y-coordinates (locations.y) and design matrix (locations.F)
%             of locations where prediction is desired
%
% data........the data set as a structure array of the format data.x (x-coordinates), data.y
%             (y-coordinates), data.z (measured values)
%
% model.......contains the parameters of the spatial copula model (see
%             function initialize)
%
% search......search radius for interpolation
%
% calc........structure array indicating which predictive quantities to compute
%             calc.mean.....either 1 or 0 depending on whether to compute
%                           the mean of the predictive distribution or not
%             calc.std.....either 1 or 0 depending on whether to compute
%                           the standard deviation of the predictive 
%                           distribution or not
%             calc.quantiles....vector of quantiles to compute (empty if no
%                               quantiles are sought)
%
% options.....options structure array (see function testOptions)
%
% Outputs:
%
% prediction...structure array consisting of the prediction results
%              prediction.mean...vector of predictive means
%              prediction.std....vector of predictive standard deviations
%              prediction.quantiles...matrix (number of locations x number
%                                     of quantiles) consisting of predictive 
%                                     quantiles
%
% Hannes Kazianka, August 2011

    
    options=testOptions(options);
    
    meanValue=NaN.*zeros(length(locations.x),1);
    varValue=NaN.*zeros(length(locations.x),1);
    if ~isempty(calc.quantiles)
        quantilesValue=NaN.*zeros(length(locations.x),length(calc.quantiles));
    else
        quantilesValue=NaN.*zeros(length(locations.x),1);
    end
    
    if ~isempty(model.anisotropy)
        xy=[cos(model.anisotropy.params(1)),sin(model.anisotropy.params(1));-model.anisotropy.params(2)*sin(model.anisotropy.params(1)),model.anisotropy.params(2)*cos(model.anisotropy.params(1))]*[data.x';data.y'];
        data.x=xy(1,:)';
        data.y=xy(2,:)';        
        xy=[cos(model.anisotropy.params(1)),sin(model.anisotropy.params(1));-model.anisotropy.params(2)*sin(model.anisotropy.params(1)),model.anisotropy.params(2)*cos(model.anisotropy.params(1))]*[locations.x';locations.y'];
        locations.x=xy(1,:)';
        locations.y=xy(2,:)';
    end

    expected=model.trend.F*model.trend.params';
    if length(model.margin.params)==1
        newdata=feval([model.margin.name 'cdf'],data.z,expected,model.margin.params); 
    elseif length(model.margin.params)==2
        newdata=feval([model.margin.name 'cdf'],data.z,expected,model.margin.params(1),model.margin.params(2));
    end
    
    multeps=[];
    if isempty(model.copula) || strcmp(model.copula.name,'Gaussian')
        newdata=norminv(newdata,0,1);
    else
        switch model.copula.name
            case 'Chi2'
                ijmatrix=zeros(2^(search+1),search+1);
                for j=1:(search+1)
                    ijmatrix(:,j)=kron(mod(0:(2^(search-j+2)-1),2)',ones(2^(j-1),1));
                end
                multeps=(-1).^ijmatrix;
                newdata=ncx2inv(newdata,1,model.copula.params);
            case 'Student'
                newdata=tinv(newdata,model.copula.params);
        end
    end    
    
    h=pdist([data.x data.y]);
    h=squareform(h);  
    
     if options.parallel==1
        if matlabpool('size')==0
            matlabpool
        end
        options.parallel=0;
        parfor i=1:length(locations.x)
            [meanValue(i),varValue(i),quantilesValue(i,:)]=copulaPredictLoop(i,locations,data,model,search,calc,options,newdata,h,multeps);
        end
    else
        for i=1:length(locations.x)
            [meanValue(i),varValue(i),quantilesValue(i,:)]=copulaPredictLoop(i,locations,data,model,search,calc,options,newdata,h,multeps);
        end
     end
    
    prediction=[];
    if  calc.mean==1
        prediction.mean=meanValue;
    end
    if calc.std==1
        prediction.std=sqrt(varValue);
    end
    if ~isempty(calc.quantiles)
        prediction.quantiles=quantilesValue;
    end
    
function [meanValue varValue quantilesValue]=copulaPredictLoop(i,locations,data,model,search,calc,options,newdata,h,multeps)
        
        meanValue=NaN;
        varValue=NaN;
        quantilesValue=NaN;
        if ~isnan(locations.x(i))
            distances=sqrt((locations.x(i)-data.x).^2+(locations.y(i)-data.y).^2);
            [distances index]=sort(distances);
            distances=distances(1:(search))';
            index=index(1:(search));
            F=locations.F(i,:)*model.trend.params';
            sigma=correlCovMat([0 distances;[distances' h(index,index)]],model.correlation.params,model.correlation.model);
            
            
            if calc.mean==1
                if isempty(model.copula) || strcmp(model.copula.name,'Gaussian')
                    %sigma22inv=inv(sigma(2:end,2:end));
                    %m=sigma(1,2:end)*sigma22inv*newdata(index);
                    %v=sqrt(1-sigma(1,2:end)*sigma22inv*sigma(2:end,1));
                    temp=sigma(1,2:end)/sigma(2:end,2:end);
                    m=temp*newdata(index);
                    v=sqrt(1-temp*sigma(2:end,1));
                    meanValue=quad(@meanintegrateGaussian,0,1,1e-9,0,m,v,model.margin,F);
                else
                    switch model.copula.name
                        case 'Chi2'
                            sigmaChol=chol(sigma,'upper');
                            sigma22Chol=chol(sigma(2:end,2:end),'upper');
                            eps=multeps(1:2^search,1:search).*kron(ones(2^search,1),sqrt(newdata(index))');
                            c=sum(mvnpdf_noinvert(eps,sqrt(model.copula.params)*ones(1,search),sigma22Chol));
                            meanValue=quad(@meanintegrateChi2,0,1,1e-5,0,model.copula.params,sigmaChol,multeps,newdata(index),search,model.margin,F,c);
                        case 'Student'
                            %sigma22inv=inv(sigma(2:end,2:end));
                            %m=sigma(1,2:end)*sigma22inv*newdata(index);
                            %v=sqrt(copula.params/(copula.params+search)*(1+newdata(index)'*sigma22inv*newdata(index)/copula.params)*(1-sigma(1,2:end)*sigma22inv*sigma(2:end,1)));
                            temp=sigma(1,2:end)/sigma(2:end,2:end);
                            m=temp*newdata(index);
                            v=sqrt(model.copula.params/(model.copula.params+search)*(1+newdata(index)'*(sigma(2:end,2:end)\newdata(index))/model.copula.params)*(1-temp*sigma(2:end,1)));
                            meanValue=quad(@meanintegrateStudent,0,1,1e-9,0,m,v,model.margin,F,model.copula.params,search);
                    end
                end
            end
            
            if calc.std==1 && calc.mean==1
                if isempty(model.copula) || strcmp(model.copula.name,'Gaussian')
                	varValue=quad(@varintegrateGaussian,0,1,1e-9,0,m,v,model.margin,F,meanValue);
                else
                    switch model.copula.name
                        case 'Chi2'
                            varValue=quad(@varintegrateChi2,0,1,1e-5,0,model.copula.params,sigmaChol,multeps,newdata(index),search,model.margin,F,meanValue,c);
                        case 'Student'
                            varValue=quad(@varintegrateStudent,0,1,1e-9,0,m,v,model.margin,F,meanValue,model.copula.params,search);%%
                    end
                end
            end
            
            if ~isempty(calc.quantiles) && (isempty(model.copula) || strcmp(model.copula.name,'Gaussian') || strcmp(model.copula.name,'Student'))                   
                test=0.1:0.1:0.9;
                integr=zeros(length(test),1);
                quantilesValue=zeros(1,length(calc.quantiles));
                if isempty(model.copula) || strcmp(model.copula.name,'Gaussian')
                    for zz=1:length(test)
                        integr(zz)=quad(@condcopuladensGaussian,0,test(zz),1e-5,0,m,v);
                    end
                else
                    switch model.copula.name
                        case 'Student'
                            for zz=1:length(test)
                                integr(zz)=quad(@condcopuladensStudent,0,test(zz),1e-5,0,m,v,model.copula.params,search);%%
                            end 
                    end
                end
                for j=1:length(calc.quantiles)
                    for zz=1:length(test)
                        val(zz)=calc.quantiles(j)-integr(zz);
                    end
                    zz=find(val.^2==min(val.^2));
                    q=test(zz);
						
                    if length(test)==1
                        if val(zz)>=0
                            q1=q;
                            f1=val(zz);
                            q2=0.9999999;
                            f2=calc.quantiles(j)-1;
                        else
                            q1=0.0000001;
                            f1=calc.quantiles(j);
                            q2=q;
                            f2=val(zz);
                        end
                    elseif zz~=length(test) && zz~=1
                        if val(zz)>=0
                            q1=q;	
                            f1=val(zz);
                            q2=test(zz+1);
                            f2=val(zz+1);
                        else
                            q1=test(zz-1);
                            f1=val(zz-1);
                            q2=q;
                            f2=val(zz);
                        end
                    elseif zz==1
                        if val(zz)>=0
                            q1=q;
                            f1=val(zz);
                            q2=test(zz+1);
                            f2=val(zz+1);
                        else 
                            q1=0.0000001;
                            f1=calc.quantiles(j);
                            q2=q;
                            f2=val(zz);
                        end
                    else
                        if val(zz)>=0
                            q1=q;
                            f1=val(zz);
                            q2=0.9999999;
                            f2=calc.quantiles(j)-1;
                        else
                            q1=test(zz-1);
                            f1=val(zz-1);
                            q2=q;
                            f2=val(zz);
                        end
                    end

                    if sign(f1)~=-sign(f2)
                        q1=0;
                        q2=1;
                    end

                    if isempty(model.copula) || strcmp(model.copula.name,'Gaussian')
                        [quantilesValue(j) val]=fzero(@quantilesGaussian,[q1,q2],[],m,v,calc.quantiles(j));
                    else 
                        switch model.copula.name
                            case 'Student'
                                [quantilesValue(j) val]=fzero(@quantilesStudent,[q1,q2],[],m,v,calc.quantiles(j),model.copula.params,search);%%
                        end
                    end
                end
                    
                if length(model.margin.params)==1
                    quantilesValue=feval([model.margin.name 'inv'],quantilesValue,F,model.margin.params); 
                elseif length(model.margin.params)==2
                    quantilesValue=feval([model.margin.name 'inv'],quantilesValue,F,model.margin.params(1),model.margin.params(2));
                end
            end
        end
        if options.print>=1
            disp(['Prediction: Location ' num2str(i)]);
        end

%%
%GAUSSIAN COPULA

function val = quantilesGaussian(loc,m,v,quant)    
    if loc<1 && loc >0
        val=quant-quad(@condcopuladensGaussian,0,loc,1e-9,0,m,v);
    else 
        val=1e10;
    end
      


function d=meanintegrateGaussian(loc,m,v,margin,expected)
    if length(margin.params)==1
        locx=feval([margin.name 'inv'],loc,expected,margin.params); 
    elseif length(margin.params)==2
        locx=feval([margin.name 'inv'],loc,expected,margin.params(1),margin.params(2));
    end    
    d=locx.*condcopuladensGaussian(loc,m,v);
    if sum(isnan(d))>0
        d(isnan(d))=0;
    end


function d=varintegrateGaussian(loc,m,v,margin,expected,mean1)
    if length(margin.params)==1
        locx=feval([margin.name 'inv'],loc,expected,margin.params); 
    elseif length(margin.params)==2
        locx=feval([margin.name 'inv'],loc,expected,margin.params(1),margin.params(2));
    end
    d=(locx-mean1).^2.*condcopuladensGaussian(loc,m,v);
    if sum(isnan(d))>0
        d(isnan(d))=0;
    end


function distrib=condcopuladensGaussian(loc,m,v)
    loc=norminv(loc,0,1);
    div=normpdf(loc,0,1);
    distrib=normpdf(loc,m,v)./div;


%%
%STUDENT COPULA
function val = quantilesStudent(loc,m,v,quant,nu,search)    
    if loc<1 && loc >0
        val=quant-quad(@condcopuladensStudent,0,loc,1e-9,0,m,v,nu,search);
    else 
        val=1e10;
    end
      


function d=meanintegrateStudent(loc,m,v,margin,expected,nu,search)
    if length(margin.params)==1
        locx=feval([margin.name 'inv'],loc,expected,margin.params); 
    elseif length(margin.params)==2
        locx=feval([margin.name 'inv'],loc,expected,margin.params(1),margin.params(2));
    end    
    d=locx.*condcopuladensStudent(loc,m,v,nu,search);
    if sum(isnan(d))>0
        d(isnan(d))=0;
    end


function d=varintegrateStudent(loc,m,v,margin,expected,mean1,nu,search)
    if length(margin.params)==1
        locx=feval([margin.name 'inv'],loc,expected,margin.params); 
    elseif length(margin.params)==2
        locx=feval([margin.name 'inv'],loc,expected,margin.params(1),margin.params(2));
    end
    d=(locx-mean1).^2.*condcopuladensStudent(loc,m,v,nu,search);
    if sum(isnan(d))>0
        d(isnan(d))=0;
    end


function distrib=condcopuladensStudent(loc,m,v,nu,search)
    loc=tinv(loc,nu);
    div=tpdf(loc,nu);
    distrib=(tpdf((loc-m)./v,nu+search)./div)./v;
    
        
%%
%CHI^2-COPULA

function distrib=condcopuladensChi2(loc,lambda,sigmaChol,multeps,newdata,search,c)
    loc=ncx2inv(loc,1,lambda);
    div=ncx2pdf(loc,1,lambda);

    distrib=zeros(length(loc),1);
    for i=1:length(loc)
        eps=multeps.*kron(ones(2^(search+1),1),[sqrt(loc(i));sqrt(newdata)]');
        distrib(i)=sum(mvnpdf_noinvert(eps,sqrt(lambda)*ones(1,search+1),sigmaChol))/(2*sqrt(loc(i))*div(i)*c);
    end


function d=meanintegrateChi2(loc,lambda,sigmaChol,multeps,newdata,search,margin,expected,c)
    if length(margin.params)==1
        locx=feval([margin.name 'inv'],loc,expected,margin.params); 
    elseif length(margin.params)==2
        locx=feval([margin.name 'inv'],loc,expected,margin.params(1),margin.params(2));
    end

    d=locx.*condcopuladensChi2(loc,lambda,sigmaChol,multeps,newdata,search,c)';
    if sum(isnan(d))>0
        d(isnan(d))=0;
    end


function d=varintegrateChi2(loc,lambda,sigmaChol,multeps,newdata,search,margin,expected,mean1,c)
    if length(margin.params)==1
        locx=feval([margin.name 'inv'],loc,expected,margin.params); 
    elseif length(margin.params)==2
        locx=feval([margin.name 'inv'],loc,expected,margin.params(1),margin.params(2));
    end
    d=(locx-mean1).^2.*condcopuladensChi2(loc,lambda,sigmaChol,multeps,newdata,search,c)';
    if sum(isnan(d))>0
        d(isnan(d))=0;
    end








