function dens= copulaPredictDensityDiscreteGQT(locations,data,model,search,grid,options)

% Computes the predictive density at given locations for discrete data
% using the GQT method
%
% Inputs:
%
% locations...structure array representing the x-coordinates (locations.x),
%             y-coordinates (locations.y) and design matrix (locations.F)
%             of locations where prediction is desired
%
% data........the data set as a structure array of the format data.x (x-coordinates), data.y
%             (y-coordinates), data.z (measured values)
%
% model.......contains the parameters of the spatial copula model (see
%             function initialize)
%
% search......search radius for interpolation
%
% grid........the grid on which the density is computed (discretization of
%             the support of the predictive distribution)
%
% options.....options structure array (see function testOptions)
%
% Outputs:
%
% dens........matrix (number of observations x grid size) consisting of the
%             predictive densities
%
% Hannes Kazianka, August 2012

options=testOptions(options);

if ~isempty(model.anisotropy)
    xy=[cos(model.anisotropy.params(1)),sin(model.anisotropy.params(1));-model.anisotropy.params(2)*sin(model.anisotropy.params(1)),model.anisotropy.params(2)*cos(model.anisotropy.params(1))]*[data.x';data.y'];
    data.x=xy(1,:)';
    data.y=xy(2,:)';        
    xy=[cos(model.anisotropy.params(1)),sin(model.anisotropy.params(1));-model.anisotropy.params(2)*sin(model.anisotropy.params(1)),model.anisotropy.params(2)*cos(model.anisotropy.params(1))]*[locations.x';locations.y'];
    locations.x=xy(1,:)';
    locations.y=xy(2,:)';
end


expected=model.trend.F*model.trend.params';
if strcmp(model.margin.name,'poiss')
	expected=exp(expected);
	zz=poisspdf(data.z,expected); 
	b=poisscdf(data.z-1,expected); 
elseif strcmp(model.margin.name,'bino')
    expected=exp(expected)./(1+exp(expected));
    zz=binopdf(data.z,model.margin.paramsfix,expected);% paramsfix=different values of n_i possible
    b=binocdf(data.z-1,model.margin.paramsfix,expected);
elseif strcmp(model.margin.name,'geo')
    expected=exp(expected)./(1+exp(expected));
    zz=geopdf(data.z,expected);
    b=geocdf(data.z-1,expected);
end
newdata=(2*b+zz)/2;

multeps=[];
if isempty(model.copula) || strcmp(model.copula.name,'Gaussian')
    newdata=norminv(newdata,0,1);
else
    switch model.copula.name
        case 'Chi2'
            ijmatrix=zeros(2^(search+1),search+1);
            for j=1:(search+1)
                ijmatrix(:,j)=kron(mod(0:(2^(search-j+2)-1),2)',ones(2^(j-1),1));
            end
            multeps=(-1).^ijmatrix;
            newdata=ncx2inv(newdata,1,model.copula.params);
        case 'Student'
            newdata=tinv(newdata,model.copula.params);
    end
end 

h=pdist([data.x data.y]);
h=squareform(h);  
dens=NaN.*zeros(length(locations.x),length(grid));

if options.parallel==1
    if matlabpool('size')==0
        matlabpool
    end
    options.parallel=0;
    parfor i=1:length(locations.x)
        dens(i,:)=copulaPredictDensityDiscreteGQTLoop(i,locations,data,model,search,grid,options,newdata,h,multeps);
    end
else
    for i=1:length(locations.x)   
        dens(i,:)=copulaPredictDensityDiscreteGQTLoop(i,locations,data,model,search,grid,options,newdata,h,multeps);
    end
end


function dens=copulaPredictDensityDiscreteGQTLoop(i,locations,data,model,search,grid,options,newdata,h,multeps)
	
if ~isnan(locations.x(i))

    expectedLoc=locations.F(i,:)*model.trend.params';
    if strcmp(model.margin.name,'poiss')
        expectedLoc=exp(expectedLoc);
        zzLoc=poisspdf(grid,expectedLoc); 
        bLoc=poisscdf(grid-1,expectedLoc); 
    elseif strcmp(model.margin.name,'bino')
            expectedLoc=exp(expectedLoc)./(1+exp(expectedLoc));
            zzLoc=binopdf(grid,locations.paramsfix(i),expectedLoc); %paramsfix=verschiedene n_i
            bLoc=binocdf(grid-1,locations.paramsfix(i),expectedLoc);
    elseif strcmp(model.margin.name,'geo')
            expectedLoc=exp(expectedLoc)./(1+exp(expectedLoc));
            zzLoc=geopdf(grid,expectedLoc);
            bLoc=geocdf(grid-1,expectedLoc);
    end
    gridx=(2*bLoc+zzLoc)/2;
        
    if isempty(model.copula) || strcmp(model.copula.name,'Gaussian')
                gridx=norminv(gridx);
                div=normpdf(gridx);
    else
        switch model.copula.name
            case 'Chi2'
                gridx=ncx2inv(gridx,1,model.copula.params);
                div=ncx2pdf(gridx,1,model.copula.params);
            case 'Student'
                gridx=tinv(gridx,model.copula.params);
                div=tpdf(gridx,model.copula.params);
        end
    end
    
    distances=sqrt((locations.x(i)-data.x).^2+(locations.y(i)-data.y).^2);
    [distances index]=sort(distances);
    distances=distances(1:(search))';
    index=index(1:(search));
            
    covi=correlCovMat([0 distances;[distances' h(index,index)]],model.correlation.params,model.correlation.model);
    if isempty(model.copula) || strcmp(model.copula.name,'Gaussian')
        %sigma22inv=inv(covi(2:end,2:end));
        u=covi(1,2:end)/covi(2:end,2:end);%u=covi(1,2:end)*sigma22inv;
        dens=(normpdf(gridx,u*newdata(index),sqrt(1-u*covi(2:end,1)))./div).*zzLoc;
    else
        switch model.copula.name
            case 'Chi2'
                dens=zeros(1,length(grid));
                newdataindex=sqrt(newdata(index));
                eps=multeps(1:2^search,1:search).*kron(ones(2^search,1),newdataindex');
                c=sum(mvnpdf(eps,sqrt(model.copula.params)*ones(1,search),covi(2:end,2:end)));
                R = chol(covi,'upper');
                d=size(covi,1);
                logSqrtDetSigma = sum(log(diag(R)));
                mu=repmat(sqrt(model.copula.params)*ones(1,search+1),size(multeps,1),1);
                Rinv=eye(size(covi)) / R;
                for j=1:length(grid)
                    eps=multeps.*kron(ones(2^(search+1),1),[sqrt(gridx(j));newdataindex]');
                    xRinv=(eps - mu)*Rinv;
                    quadform = sum(xRinv.^2, 2);
                    dens(j)=sum(exp(-0.5*quadform - logSqrtDetSigma - d*log(2*pi)/2))/(2*sqrt(gridx(j))*div(j)*c);
                end
                dens=dens.*zzLoc;
            case 'Student'
                %sigma22inv=inv(covi(2:end,2:end));
                u=covi(1,2:end)/covi(2:end,2:end);
                m=u*newdata(index);%m=covi(1,2:end)*sigma22inv*newdata(index);
                %v=sqrt(copula.params/(copula.params+search)*(1+newdata(index)'*sigma22inv*newdata(index)/copula.params)*(1-covi(1,2:end)*sigma22inv*covi(2:end,1)));
                v=sqrt(model.copula.params/(model.copula.params+search)*(1+newdata(index)'*(covi(2:end,2:end)\newdata(index))/model.copula.params)*(1-u*covi(2:end,1)));
                dens=((tpdf((gridx-m)./v,model.copula.params+search)./div)./v).*zzLoc;
        end
    end
            
end
if options.print>=1
    disp(['Prediction: Location ' num2str(i)]);
end
    
    
