function result=copulaTest(data,model,n,lag,alpha,options)

% Performs parametric bootstrap-based goodness-of-fit test.
%
% Inputs:
%
% data........the data set as a structure array of the format data.x (x-coordinates), data.y
%             (y-coordinates), data.z (measured values)
%
% model.......contains the parameters of the spatial copula model (see
%             function initialize)
%
% n...........number of bootstrap iterations
%
% lag.........lag classes to test
%
% alpha.......significance level
%
% options.....options from testOptions
% 
%
% Outputs:
%
% result......structure array with the test results
%             result.D1normorig.....Cramer-von Mises test statistics for the original data
%             result.D2normorig.....Kolmogorov-Smirnov test statistics for the oroginal data
%             result.D1norm.........Cramer-von Mises test statistics for the simulated data
%             result.D2norm.........Kolmogorov-Smirnov test statistics for the simulated data
%             result.paramorig......original correlation function parameters
%             result.param..........correlation function parameters estimated from the simulated data
%             result.numberD1norm...number of CM-test statistics smaller or equal than the original MS-test statistic
%             result.numberD2norm...number of KS-test statistics smaller or equal than the original KS-test statistic
%             result.resultD1norm...null hypothesis accepted or rejected by the CM-test
%             result.resultD2norm...null hypothesis accepted or rejected by the KS-test
%             
%  
% Hannes Kazianka, August 2012


    options=testOptions(options);

    if ~isempty(model.anisotropy.params)
        xy=[cos(model.anisotropy.params(1)),sin(model.anisotropy.params(1));-model.anisotropy.params(2)*sin(model.anisotropy.params(1)),model.anisotropy.params(2)*cos(model.anisotropy.params(1))]*[data.x';data.y'];
        data.x=xy(1,:)';
        data.y=xy(2,:)';
    end
    
    expected=model.trend.F*model.trend.params';
    if length(model.margin.params)==1
        data.z=feval([model.margin.name 'cdf'],data.z,expected,model.margin.params); 
    elseif length(model.margin.params)==2
        data.z=feval([model.margin.name 'cdf'],data.z,expected,model.margin.params(1),model.margin.params(2));
    end

    distancematrix=pdist([data.x data.y]);
    distancematrix=squareform(distancematrix);

    [a b]=meshgrid(0.01:0.02:0.99,0.01:0.02:0.99);
    transfgrid(:,1)=a(:);
    transfgrid(:,2)=b(:);
    clear a b
    
    if isempty(model.copula) || strcmp(model.copula.name,'Gaussian')
        transfgridinv=norminv(transfgrid);
    else
        switch model.copula.name
            case 'Chi2'
                transfgridinv=ncx2inv(transfgrid,1,model.copula.params);
                transfgridinv=sqrt(transfgridinv);
            case 'Student'
                transfgridinv=tinv(transfgrid,model.copula.params);
        end
    end
    
    D1normorig=zeros(length(lag)-1,1);
    D2normorig=zeros(length(lag)-1,1);
    empcopulaorig=empiricalCopula(data,lag,transfgrid,distancematrix);
    for h=1:(length(lag)-1)
        cov=correlCovMat([0 (lag(h)+lag(h+1))/2;(lag(h)+lag(h+1))/2 0],model.correlation.params,model.correlation.model);
        
        if isempty(model.copula) || strcmp(model.copula.name,'Gaussian')
            theoreticalcopula=normcopula(transfgridinv(:,1),transfgridinv(:,2),cov);
        else
            switch model.copula.name
                case 'Chi2'
                    theoreticalcopula=chi2copula(transfgridinv(:,1),transfgridinv(:,2),cov,sqrt(model.copula.params));
                case 'Student'
                    theoreticalcopula=studentcopula(transfgridinv(:,1),transfgridinv(:,2),cov,model.copula.params);
            end
        end
        
        D1normorig(h)=sum((theoreticalcopula-empcopulaorig{h}).^2);
        D2normorig(h)=max(abs(theoreticalcopula-empcopulaorig{h}));
    end

    D1norm=zeros(n,(length(lag)-1));
    D2norm=zeros(n,(length(lag)-1));
    correlationHatn=zeros(n,length(model.correlation.params));
    cov=correlCovMat(distancematrix,model.correlation.params,model.correlation.model);
    
    if options.parallel==1
        if matlabpool('size')==0
            matlabpool
        end
        options.parallel=0;
        parfor i=1:n
            [D1norm(i,:) D2norm(i,:) correlationHatn(i,:)]=copulaTestLoop(i,model,data,cov,lag,transfgrid,transfgridinv,distancematrix,options);
        end
    else
        for i=1:n
            [D1norm(i,:) D2norm(i,:) correlationHatn(i,:)]=copulaTestLoop(i,model,data,cov,lag,transfgrid,transfgridinv,distancematrix,options);
        end
    end

    result.D1norm=D1norm;
    result.D2norm=D2norm;
    result.D1normorig=D1normorig;
    result.D2normorig=D2normorig;
    result.paramorig=model.correlation.params;
    result.param=correlationHatn;

    for h=1:(length(lag)-1)

        result.numberD1norm(h)=sum(D1norm(:,h)<=D1normorig(h));
        result.numberD2norm(h)=sum(D2norm(:,h)<=D2normorig(h));
        
        if result.numberD1norm(h)>(n-alpha*n) 
            result.resultD1norm{h}='negative';
        else
            result.resultD1norm{h}='positive';
        end
        if result.numberD2norm(h)>(n-alpha*n) 
            result.resultD2norm{h}='negative';
        else
            result.resultD2norm{h}='positive';
        end
    end

    
function [D1normi D2normi correlationHati]=copulaTestLoop(i,model,field,cov,lag,transfgrid,transfgridinv,distancematrix,options)

    D1normi=zeros(1,(length(lag)-1));
    D2normi=zeros(1,(length(lag)-1));

    if isempty(model.copula) || strcmp(model.copula.name,'Gaussian')
        field.z=mvnrnd(zeros(1,size(cov,1)),cov,1)';
        correlationHati=fmincon(@likeliGaussian,model.correlation.params,[],[],[],[],model.correlation.lower,model.correlation.upper,[],options.optim,field.z,model.correlation,distancematrix);
    else 
        switch model.copula.name
            case 'Chi2'
                field.z=mvnrnd(ones(size(cov,1),1).*sqrt(model.copula.params),cov,1)';
                field.z=field.z.^2;
                correlationHati=fmincon(@likeliChi2,model.correlation.params,[],[],[],[],model.correlation.lower,model.correlation.upper,[],options.optim,sqrt(field.z),model.correlation,distancematrix,sqrt(model.copula.params));
            case 'Student'
                field.z=mvtrnd(cov,model.copula.params,1)';
                correlationHati=fmincon(@likeliStudent,model.correlation.params,[],[],[],[],model.correlation.lower,model.correlation.upper,[],options.optim,field.z,model.correlation,distancematrix,model.copula.params);
        end
    end

    empcopula=empiricalCopula(field,lag,transfgrid,distancematrix);
    clear cov

    for h=1:(length(lag)-1)
        covTemp=correlCovMat([0 (lag(h)+lag(h+1))/2;(lag(h)+lag(h+1))/2 0],correlationHati,model.correlation.model);
        if isempty(model.copula) || strcmp(model.copula.name,'Gaussian')
            theoreticalcopula=normcopula(transfgridinv(:,1),transfgridinv(:,2),covTemp);
        else
            switch model.copula.name
                case 'Chi2'
                    theoreticalcopula=chi2copula(transfgridinv(:,1),transfgridinv(:,2),covTemp,sqrt(model.copula.params));
                case 'Student'
                    theoreticalcopula=studentcopula(transfgridinv(:,1),transfgridinv(:,2),covTemp,model.copula.params);
            end
        end

        D1normi(h)=sum(sum((theoreticalcopula-empcopula{h}).^2));
        D2normi(h)=max(max(abs(theoreticalcopula-empcopula{h})));
    end
    if options.print>=1
        disp(['Simulation ' num2str(i)]);
    end
%%
%GAUSSIAN COPULA

function like = likeliGaussian(param,z,correlation,h)
    cov=correlCovMat(h,param,correlation.model);
    like=-mvnlogpdf_noconst(z',cov);
    if isnan(like)
        like=1e100;
    end

function val = normcopula(x,y,cov)
    val=mvncdf([x,y],[0,0],cov);

%%
%CHI^2-COPULA   

function like = likeliChi2(param,sqrtz,correlation,h,sqrtlambda)
    cov=correlCovMat(h,param,correlation.model);
    like=0;
    logsqrtz=log(sqrtz);
    m=[sqrtlambda,sqrtlambda];
    for i=1:length(sqrtz)
        for j=(i+1):length(sqrtz)
            like=like-(log(sum(mvnpdf([sqrtz(i),sqrtz(j);-sqrtz(i),sqrtz(j);sqrtz(i),-sqrtz(j);-sqrtz(i),-sqrtz(j)],m,cov([i,j],[i,j]))))-logsqrtz(i)-logsqrtz(j));
        end
    end
    
    if isnan(like) || isinf(like)
        like=1e100;
    end

  

function val = chi2copula(sqrtx,sqrty,cov,sqrtlambda)
    m=[sqrtlambda sqrtlambda];
    val=mvncdf([sqrtx,sqrty],m,cov)-mvncdf([-sqrtx,sqrty],m,cov)-mvncdf([sqrtx,-sqrty],m,cov)+mvncdf([-sqrtx,-sqrty],m,cov);

%%
%STUDENT-COPULA

function like = likeliStudent(param,z,correlation,h,nu)
    cov=correlCovMat(h,param,correlation.model);
    n=length(z);
    R = chol(cov,'upper');
    xRinv = z' / R;
    logSqrtDetSigma = sum(log(diag(R)));
    quadform = sum(xRinv.^2, 2);
    like=gammaln(nu/2)-gammaln((nu+n)/2)+n/2*log(nu)+0.5*(nu+n)*log(1+1/nu*quadform)+logSqrtDetSigma;
    
    if isnan(like)
        like=1e100;
    end

function val = studentcopula(x,y,cov,nu)
    val=mvtcdf([x,y],cov,nu);
