function cov=correlCovMat(h,delta,model,sigma2)
    
% Computes the correlation (covariance) matrix for given distance matrix 
% and correlation model. 
%
% Inputs:
%
% h........distance matrix
%
% delta....correlation function parameters in the following order: nugget,
%          range, possible other (smoothness, mixture...) parameters
%
% model....on of the following correlation models:
%          'Gau'...Gaussian
%          'Exp'...Exponential
%          'Mix'...a mixture of Gaussian and Exponential model, same range
%          'Sph'...Spherical
%          'Mat'...Matern model using M. Stein's representation
%          'Mix2'...a mixture of Gaussian and Exponential model, different ranges
%          'Pow'...power exponential
%
% sigma2...a possible variance parameter
%
% Outputs:
%
% cov.....the correlation (covariance) matrix
%
% Hannes Kazianka, August 2011

    if nargin<4
        sigma2=1;
    end

    switch model
        case 'Gau'
            cov=(1-delta(1)).*exp(-(h.^2)./(delta(2)^2));
        case 'Exp'
            cov=(1-delta(1)).*exp(-h./delta(2));
        case 'Mix'
            cov=(1-delta(1)).*((1-delta(3)).*exp(-h./delta(2))+delta(3).*exp(-(h.^2)./(delta(2)^2)));
        case 'Sph'
            cov=(1-delta(1))*(1-(3.*h./(2*delta(2))-h.^3./(2*delta(2)^3)));
            cov(h>delta(2))=0;
        case 'Mat'
            if size(h,1)==size(h,2) && size(h,1)>1
                u=ones(size(h));
                hnew=h(logical(tril(u,-1)));
                matern=besselk(delta(3),2*delta(3)^(1/2).*hnew./delta(2));
                multipl=(1-delta(1))/(2^(delta(3) - 1)*gamma(delta(3)))*(2*delta(3)^(1/2)*hnew./delta(2)).^delta(3);
                covnew=matern.*multipl;
                covnew(matern==0 | isinf(multipl))=0;
                covnew(isinf(matern))=1-delta(1);
                cov=zeros(size(h));
                cov(logical(tril(u,-1)))=covnew;
                cov=cov+cov'+eye(size(h,1));
            else
                matern=besselk(delta(3),2*delta(3)^(1/2).*h./delta(2));
                multipl=(1-delta(1))/(2^(delta(3) - 1)*gamma(delta(3)))*(2*delta(3)^(1/2)*h./delta(2)).^delta(3);
                cov=matern.*multipl;
                cov(matern==0 | isinf(multipl))=0;
                cov(isinf(matern))=1-delta(1);
            end
        case 'Mix2'
            cov=(1-delta(1)).*((1-delta(4)).*exp(-h./delta(2))+delta(4).*exp(-(h.^2)./(delta(3)^2)));
        case 'Pow'
            cov=(1-delta(1)).*exp(-(h./delta(2)).^(delta(3)));
    end
    cov(h==0)=1;
    cov=cov.*sigma2;
    