function [copula Sh] =empiricalCopula(data,lags,prob,h)

% Computes the empirical copula for certain lags.
%
% Inputs:
%
% data........the data set as a structure array of the format data.x (x-coordinates), data.y
%             (y-coordinates), data.z (measured values)
%
% lags........lag classes to investigate
%
% prob........vector of probabilities for which P(X_1\leq p_i,X_2\leq p_j)
%             is sought
%
% h...........distance matrix if available (otherwise computed from data)
%
% Outputs:
%
% copula......structure array of empirical copulas at the different lags
%
% Sh..........structure array of rank transformed data pairs at the lags
%
% Hannes Kazianka, August 2011


    if nargin<4
        h=pdist([data.x data.y]);
        h=squareform(h);
    end
    newdata=sort(data.z);
    distrib=cumsum(ones(1,length(newdata)))./(length(newdata)+1);
    for i=1:length(newdata)
        data.z(i)=empdistrib(newdata,data.z(i),distrib);
    end
    Sh=cell(length(lags)-1,1);
    copula=cell(length(lags)-1,1);
    for k=1:(length(lags)-1)
        index=h>lags(k) & h<=lags(k+1);
        Z1=kron(ones(length(data.z),1),data.z').*index;
        UD=triu(Z1);
        LD=tril(Z1)';
        UD=UD(UD~=0);
        LD=LD(LD~=0);
        Sh{k}=[UD LD;LD UD];
    end
    if nargin<3
        for k=1:(length(lags)-1)
            copula{k}=zeros(size(Sh{k},1),1);
            for i=1:size(Sh{k},1)
                copula{k}(i)=(1/size(Sh{k},1))*sum((Sh{k}(:,1)<=Sh{k}(i,1)).*(Sh{k}(:,2)<=Sh{k}(i,2)));
            end
        end
    else
        for k=1:(length(lags)-1)
            copula{k}=zeros(length(prob));
            for i=1:length(prob)
                for j=1:length(prob)
                    copula{k}(i,j)=(1/size(Sh{k},1))*sum((Sh{k}(:,1)<=prob(i)).*(Sh{k}(:,2)<=prob(j)));
                end
            end
        end
    end


function res =empdistrib(data,x,distrib)
	ind=find(data<=x);
	if isempty(ind)==0
		res=distrib(ind(end));
    else
 		res=0;
	end