function [params do]=estimateAnisotropy(data)

% Performs anisotropy parameter estimation based on the CTI method. Only
% for >=50 observations.
%
%
% data.........the data set as a structure array of the format data.x (x-coordinates), data.y
%             (y-coordinates), data.z (measured values)
% 
% Outputs:
% 
% params.......2-vector of parameters of the geometric anisotropy             
%              params(1)....anisotropy angle >=0,<=pi
%              params(2)....anisotropy ratio >=1
%
% do...........indicates whether or not the transformation should be done
%
% Hannes Kazianka, August 2011
%
% A. Chorti and D. Hristopulos (2008). Non-parametric Identification of Anisotropic 
% (Elliptic) Correlations in Spatially Distributed Data Sets, IEEE Transactions on 
% Signal Processing, 56(10), 4738-4751 (2008). 



     
    do=0;
    params=[0 1];
    n=length(data.x);
    if n>=50
        gridsize=min([abs((max(data.x)-min(data.x))/sqrt(n)),abs((max(data.y)-min(data.y))/sqrt(n))]);
        [xi,yi] = meshgrid(min(data.x):gridsize:max(data.x),min(data.y):gridsize:max(data.y)); 
        inter=griddata(data.x,data.y,data.z,xi,yi,'linear');
        grad=gradient(inter,gridsize,gridsize);
        divX2=grad.x.^2;
        divY2=grad.y.^2;
        divXY=grad.x.*grad.y;
        Q11=mean(mean(divX2(~isnan(divX2))));
        Q22=mean(mean(divY2(~isnan(divY2))));
        Q12=mean(mean(divXY(~isnan(divXY))));
        if Q11>1e-31 && ~isnan(Q11) && ~isnan(Q12) && ~isnan(Q22)
            Zdiag=Q22/Q11;
            Zoff=Q12/Q11;
            params(1)=0.5*atan(2*Zoff/(1-Zdiag));
            params(2)=sqrt(1+((1-Zdiag)/(Zdiag-(1+Zdiag)*(cos(params(1)))^2)));
            do=testAnisotropy(numel(xi),params(2));
            if isnan(params(2))
                params=[0 1];
                do=0;
            elseif params(2)<1
                params(2)=1/params(2);
                if (params(1)+pi)>0 && params<0
                    params(1)=params(1)+pi/2;
                else
                    params(1)=params(1)-pi/2;
                end
            end
            if params(1)<0
                params(1)=pi+params(1);
            end
        end
    end
    
    
function grad = gradient(z,stepx,stepy) 
	grad.y=computeGradient(z,stepy);
	grad.x=computeGradient(z',stepx)';
        
function grad=computeGradient(z,step)
    s=size(z);
    grad=zeros(s);
    grad(1,:)=(z(2,:)-z(1,:))./step;
    grad(s(1),:)=(z(s(1),:)-z(s(1)-1,:))./step;
    grad(2:(s(1)-1),:)=(z(3:s(1),:)-z(1:(s(1)-2),:))./(2*step);
    
function do=testAnisotropy(n,ratio)
    do=1;
    r=6;
    floor=sqrt((n-2*sqrt((n-r)*r))/(n-2*r));
    ceil=sqrt((n+2*sqrt((n-r)*r))/(n-2*r));
    if ratio>floor && ratio<ceil
        do=0;
    end