function [model options]= initialize(data,options)
    
% Sets up the spatial copula model. Crude estimation of margins and
% correlation parameters to get some starting points for maximum likelihood
% estimation automatically.
% 
% Inputs:
% 
% data........the data set as a structure array of the format data.x (x-coordinates), data.y
%             (y-coordinates), data.z (measured values)
% 
% options.....options from testOptions
%
% Outputs:
% 
% model.......contains the parameters of the spatial copula model
%             model.anisotropy.params.....angle and ratio for geometric anisotropy
%             model.anisotropy.upper......upper bounds for optimization     
%             model.anisotropy.lower......lower bounds for optimization
%             model.anisotropy.const......0,1 depending on whether or not optimization should be performed
%              
%             model.trend.params.....parameters for spatial trend
%             model.trend.upper......upper bounds for optimization     
%             model.trend.lower......lower bounds for optimization
%             model.trend.F..........design matrix
%             model.trend.const......0,1 depending on whether or not optimization should be performed
%               
%             model.correlation.params.....parameters of the correlation function
%             model.correlation.model......name of the correlation model as specified in the function correlCovMat
%             model.correlation.upper......upper bounds for optimization
%             model.correlation.lower......lower bounds for optimization
%             model.correlation.const......0,1 depending on whether or not optimization should be performed 
%
%             model.margin.params.....parameters of the marginal distribution excluding the trend parameter
%             model.margin.name.......margin name e.g. 'norm', 'gevMod', 'logn', 'boxcox', 'gamMod', ...
%             model.margin.upper......upper bounds for optimization
%             model.margin.lower......lower bounds for optimization
%             model.margin.const......0,1 depending on whether or not optimization should be performed 
%
%             model.copula.params.....copula specific parameter
%             model.copula.name.......copula name: 'Gaussian', 'Chi2', 'Student'
%             model.copula.upper......upper bounds for optimization
%             model.copula.lower......lower bounds for optimization
%             model.copula.const......0,1 depending on whether or not optimization should be performed 
%
% options.....options from testOptions
%
% Hannes Kazianka, August 2011
   
    if nargin<2
        options=testOptions([]);
    else
        options=testOptions(options);
    end 
    
    copula=[];
    
    [anisotropy.params do]=estimateAnisotropy(data);
    anisotropy.lower=[0 1];
    anisotropy.upper=[pi Inf];
    anisotropy.const=0;
    
    tempData=data;
    if do==1
        xy=[cos(anisotropy.params(1)),sin(anisotropy.params(1));-anisotropy.params(2)*sin(anisotropy.params(1)),anisotropy.params(2)*cos(anisotropy.params(1))]*[data.x';data.y'];
        tempData.x=xy(1,:)';
        tempData.y=xy(2,:)';
    else
        anisotropy.params=[0 1];
    end
    
    empvario=empiricalVariogram(tempData);
    
    models=cell(0);
    models{1}='Gau';
    models{2}='Exp';
    models{3}='Sph';
    models{4}='Mat';
    MSEopt=Inf;
    
    range=0.1*sqrt((max(tempData.x) - min(tempData.x))^2 + (max(tempData.y) - min(tempData.y))^2);
    sill=mean([max(empvario.v),median(empvario.v)]);
    nugget=min(empvario.v)/sill;
    correlation=[];
    for i=1:length(models)
        temp.model=models{i};
        if strcmp(temp.model,'Mat') 
                if ~isempty(correlation.params)
                    temp.params=[correlation.params 0.5];   
                else
                    temp.params=[nugget range 0.5];
                end
                temp.upper=[0.99 Inf 10];
                temp.lower=[0.001 0.001 0.1];      
        else
            temp.params=[nugget range];
            temp.upper=[1 Inf];
            temp.lower=[0.001 0.001];

        end
        [params MSE]=fitVariogram(empvario,temp,options); 
        if MSE<MSEopt
            correlation=temp;
            correlation.params=params;
            MSEopt=MSE;
        end
    end
    correlation.const=0;
    
    [model.trend model.margin]=findmargins(data,options);
    model.correlation=correlation;
    model.copula=copula;
    model.anisotropy=anisotropy;
    
function [trend margin]=findmargins(data,options)
        
    margins=cell(0);
   if sum(data.z<=0)>0
       margins{1}='norm';
       margins{2}='gev';
   else
       margins{1}='norm';
       margins{2}='gev';
       margins{3}='logn';
       margins{4}='gam';
       margins{5}='boxcox';
   end
   
	likeOpt=-Inf;
    for i=1:length(margins)
		[params like]=eval([margins{i} 'testing(data,options);']);
        if(like>likeOpt) 
			margin.name=margins{i};
            margin.params=params(2:end);
            bounds=findmarginbounds(margin.name);
            margin.upper=bounds.upper(2:end);
            margin.lower=bounds.lower(2:end);
            trend.upper=bounds.upper(1);
            trend.lower=bounds.lower(1);
            trend.params=params(1);
            likeOpt=like;
        end
    end
    trend.F=ones(length(data.z),1);
    trend.const=0;
    margin.const=0;


function bounds = findmarginbounds(name)
	switch name
        case 'norm'
            bounds.upper=[Inf Inf];
            bounds.lower=[-Inf 0.001];
        case 'logn'
            bounds.upper=[Inf Inf];
            bounds.lower=[-Inf 0.001];
        case 'gam'
            bounds.upper=[Inf Inf];
            bounds.lower=[0.001,0.001];
        case 'gev'
            bounds.upper=[Inf Inf Inf];
            bounds.lower=[-Inf 0.001 -Inf];
        case 'boxcox'
            bounds.upper=[Inf Inf 10];
            bounds.lower=[-Inf 0.001 0.001];
	end
    
function [params like]=normtesting(data,options)
    params=[mean(data.z) std(data.z)];
    like=sum(log(normpdf(data.z,params(1),params(2))));

function [params like]=logntesting(data,options)
    params=[mean(log(data.z)) std(log(data.z))];
    like=sum(log(lognpdf(data.z,params(1),params(2))));

    function [params like]=gamtesting(data,options)
    params=gamfit(data.z);
    like=sum(log(gampdf(data.z,params(1),params(2))));
    params=[0 params];
    
function [params like]=gevtesting(data,options)
    params=gevfit(data.z);
    like=sum(log(gevpdf(data.z,params(1),params(2),params(3))));
    params=[params(3) params(2) params(1)];
    
function [params like]=boxcoxtesting(data,options)
    [lambda,like]=fmincon(@criteriumBoxCox,1,[],[],[],[],0.001,10,[],options.optim,data.z,options);
    params=[mean((data.z.^lambda-1)./lambda) std((data.z.^lambda-1)./lambda) lambda];
    like=-like;
    
function val=criteriumBoxCox(lambda,z,options)
    mu=mean((z.^lambda-1)./lambda);
    sd=std((z.^lambda-1)./lambda);
    val=-sum(log(boxcoxpdf(z,mu,sd,lambda)));
    if options.print>=2
        disp(num2str([val mu sd lambda]));
    end

    