function y = mvnpdf_noinvert(X, Mu, R)

% Computes the PDF of a multivariate Gaussian distribution without
% inversion of the covariance matrix but by using the supplied Cholesky
% factor R. Adapted from the Matlab function mvnpdf.
%
% See: mvnpdf
%
% Inputs:
%
% X......vector or matrix of values at which the PDF is sought
%
% Mu.....expected value vector or matrix
%
% R......Cholesky factor of the covariance matrix 
%
% Outputs:
%
% y......multivariate Gaussian PDF at X
%
% Hannes Kazianka, August 2011



    [n,d] = size(X);

    if numel(Mu) == 1
        X0 = X - Mu;

    elseif ndims(Mu) == 2
        [n2,d2] = size(Mu);
        if d2 ~= d
            error('stats:mvnpdf:InputSizeMismatch',...
                'X and MU must have the same number of columns.');
        elseif n2 == n 
            X0 = X - Mu;
        elseif n2 == 1 
            X0 = X - repmat(Mu,n,1);
        elseif n == 1 
            n = n2;
            X0 = repmat(X,n2,1) - Mu;
        else
            error('stats:mvnpdf:InputSizeMismatch',...
                'X or MU must be a row vector, or X and MU must have the same number of rows.');
        end
    
    else
        error('stats:mvnpdf:BadMu','MU must be a matrix.');
    end


    if ndims(R) == 2
    
        if (d == 1) && (numel(X) > 1) && (size(R,1) == n)
            X0 = X0';
            [n,d] = size(X0);
        end
    
  
        sz = size(R);
        if sz(1) ~= sz(2)
            error('stats:mvnpdf:BadCovariance',...
                'SIGMA must be a square matrix.');
        elseif ~isequal(sz, [d d])
            error('stats:mvnpdf:InputSizeMismatch',...
              'SIGMA must be a square matrix with size equal to the number of columns in X.');
        else
       
            xRinv = X0 / R;
            logSqrtDetSigma = sum(log(diag(R)));
        end
    
    end

    quadform = sum(xRinv.^2, 2);

    y = exp(-0.5*quadform - logSqrtDetSigma - d*log(2*pi)/2);
