function plotLags(data,lags,grid,model)

% Plots empirical copula and theoretical copula density for different lag
% classes on a grid
%
% Inputs:
%
% data........the data set as a structure array of the format data.x (x-coordinates), data.y
%             (y-coordinates), data.z (measured values)
%
% lags........a vector containing the different lag classes
%
% grid........grid on which the copula density is displayed
% 
% model.......either a cell array that contains a number of models, or a
%             structure array that has the format of the output of the
%             function initialize 
%  
% Hannes Kazianka, August 2011


    h=pdist([data.x data.y]);
    h=squareform(h);

	g=length(grid); 
    d=length(lags); 
    figure;
    [~,Sh]=empiricalCopula(data,lags,h);
    if nargin==4 && iscell(model)
        k=length(model);
    elseif nargin==4 && isstruct(model)
        temp=cell(0);
        temp{1}=model;
        model=temp;
        clear temp;
        k=1;
    else
        k=0;
    end
    for i=1:(d-1)
        subplot(d-1,k+1,1+(i-1)*(k+1));scatter(Sh{i}(:,1),Sh{i}(:,2));ylabel(['lag=' num2str(lags(i+1))]);
    end
    
    if k>0
        dens=zeros(g,g);
        for j=1:length(model)            
            if isempty(model{j}.copula) || strcmp(model{j}.copula.name,'Gaussian')
                gridTemp=norminv(grid);
            else 
                switch model{j}.copula.name
                    case 'Chi2'
                        gridTemp=ncx2inv(grid,1,model{j}.copula.params);
                    case 'Student'
                        gridTemp=tinv(grid,model{j}.copula.params);
                end
            end
            
            for i=1:(d-1)
                cov=correlCovMat([0 (lags(i)+lags(i+1))/2;(lags(i)+lags(i+1))/2 0],model{j}.correlation.params,model{j}.correlation.model);
                if isempty(model{j}.copula) || strcmp(model{j}.copula.name,'Gaussian')
                    for u=1:g
                        dens(u,:)=normcopuladens(gridTemp(u).*ones(g,1),gridTemp',cov);
                    end
                else 
                    switch model{j}.copula.name
                        case 'Chi2'
                            for u=1:g
                                for v=1:g
                                    dens(u,v)=chisquaredcopuladens(gridTemp(u),gridTemp(v),model{j}.copula.params,cov);
                                end
                            end
                        case 'Student'
                            for u=1:g
                                dens(u,:)=studentcopuladens(gridTemp(u).*ones(g,1),gridTemp',cov,model{j}.copula.params);
                            end
                    end
                end
                subplot(d-1,k+1,1+j+(i-1)*(k+1));imagesc(dens);set(gca,'YDir','normal');set(gca,'XTick',[1 25 50]);set(gca,'XTickLabel',[0 0.5 1]);set(gca,'YTick',[1 25 50]);set(gca,'YTickLabel',[0  0.5 1]);colorbar;
            end
        end
    end




function val=normcopuladens(u,v,cov)
    val=mvnpdf([u v],[0 0],cov)./(normpdf(u).*normpdf(v));

function val=studentcopuladens(u,v,cov,nu)
    val=mvtpdf([u v],cov,nu)./(tpdf(u,nu).*tpdf(v,nu));

function val=chisquaredcopuladens(u,v,lambda,cov)
    if lambda==0 
        x=chi2pdf(u,1);
        y=chi2pdf(v,1);
    else
        x=ncx2pdf(u,1,lambda);
        y=ncx2pdf(v,1,lambda);
    end
    
    if u~=0 && v~=0 && ~isinf(v) && ~isinf(u)
        val=sum(mvnpdf([sqrt(u) sqrt(v);-sqrt(u) sqrt(v);sqrt(u) -sqrt(v);-sqrt(u) -sqrt(v)],[sqrt(lambda) sqrt(lambda)],cov))/(4*sqrt(u*v)*x*y);
	elseif u==0 && v~=0 && ~isinf(v)
        r=cov(1,2);
        val= (1/(sqrt(1-r^2)))*cosh((sqrt(lambda*v))/(r+1))*sech(sqrt(lambda*v))*(cosh((r*(2*lambda*r+r*v-2*lambda))/(2*(r^2-1)))+sinh((r*(2*lambda*r+r*v-2*lambda))/(2*(r^2-1))));
    elseif u~=0 && v==0 && ~isinf(u)
        r=cov(1,2);
        val= (1/(sqrt(1-r^2)))*cosh((sqrt(lambda*u))/(r+1))*sech(sqrt(lambda*u))*(cosh((r*(2*lambda*r+r*u-2*lambda))/(2*(r^2-1^2)))+sinh((r*(2*lambda*r+r*u-2*lambda))/(2*(r^2-1))));
    elseif u==0 && v==0
        r=cov(1,2);
        val= exp(lambda*r/(1+r))/sqrt(1-r^2);
    elseif isinf(u) && ~isinf(v) || isinf(v) && ~isinf(u)
        val=0;
    elseif isinf(v) && isinf(u)
        val=Inf;
    end