function rho = spearmanRho(h,copula,correlation)

% Calculates the Spearman rank correlation coefficient for the vector of 
% distances h and prespecified correlation structure. If no correlation 
% structure is provided, h is taken as a vector of correlations and h<=1 is
% required.
%
% Inputs:
%
% h.............vector of distances/correlations
%
% copula........structure array of copula parameters (see the format of the
%               output of the function initialize)
%
% correlation...structure array of correlation parameters (see the format
%               of the output of the function initialize)
%
% Hannes Kazianka, August 2011
    
    if nargin<3
        ok=0;
    else
        ok=1;
    end

    for i=1:length(h)
        if ok
            cov=correlCovMat([0 h(i); h(i) 0],correlation.params,correlation.model);
        else
            cov=[1 h(i); h(i) 1];
        end
        if isempty(copula) || strcmp(copula.name,'Gaussian')
            rho(i)=12*dblquad(@copulaGaussian,0,1,0,1,1e-3,@quadl,cov)-3;
        else
            switch copula.name
                case 'Chi2'
                    rho(i)=12*dblquad(@copulaChi2,0,1,0,1,1e-3,@quadl,cov,copula.params)-3;
                case 'Student'
                    rho(i)=12*dblquad(@copulaStudent,0,1,0,1,1e-3,@quadl,cov,copula.params)-3;
            end
        end
    end



function val = copulaChi2(u,v,cov,lambda)
    if det(cov)>0.00000001
        if lambda==0 
            u=chi2inv(u,1);
            v=chi2inv(v,1);
        else
            u=ncx2inv(u,1,lambda);
            v=ncx2inv(v,1,lambda);
        end
        u=sqrt(u);
        v=sqrt(v);
        if length(u)==1 && length(v)==1
            val=mvncdf([u v;-u v;u -v;-u -v],[sqrt(lambda) sqrt(lambda)],cov)'*[1;-1;-1;1];
        else
            for i=1:length(u)
                val(i)=mvncdf([u(i) v;-u(i) v;u(i) -v;-u(i) -v],[sqrt(lambda) sqrt(lambda)],cov)'*[1;-1;-1;1];
            end
        end
    else
        val=ones(length(u),1)./3;
    end

function val=copulaGaussian(u,v,cov)
    u=norminv(u,0,1);
    v=norminv(v,0,1);
    if length(u)==1 && length(v)==1
        val=mvncdf([u v],[0 0],cov);
    else
        for i=1:length(u)
            val(i)=mvncdf([u(i) v],[0 0],cov);
        end
    end

function val=copulaStudent(u,v,cov,nu)
    u=tinv(u,nu);
    v=tinv(v,nu);
    if length(u)==1 && length(v)==1
        val=mvtcdf([u v],cov,nu);
    else
        for i=1:length(u)
            val(i)=mvtcdf([u(i) v],cov,nu);
        end
    end